import json
from typing import IO, Iterator

import httpx

from .. import errors, ty


class RawIteratorIO(IO[bytes]):
    """
    Creates a IO[bytes] from a bytes iterator.

    Adapted from https://stackoverflow.com/a/12604375/7426717
    """

    def __init__(self, it: Iterator[bytes]) -> None:
        self._it = it
        self._memory = b""

        super().__init__()

    def _read1(self, n=None) -> bytes:
        while not self._memory:
            try:
                next_memory = next(self._it)
            except StopIteration:
                break
            else:
                self._memory = next_memory

        chunk = self._memory[:n]
        self._memory = self._memory[len(chunk) :]

        return chunk

    def read(self, n=None) -> bytes:
        chunks = []
        if n is None or n < 0:
            while True:
                m = self._read1()
                if not m:
                    break
                chunks.append(m)
        else:
            while n > 0:
                m = self._read1(n)
                if not m:
                    break
                n -= len(m)
                chunks.append(m)
        return b"".join(chunks)


CHUNK_SIZE = 1024


class BaseClient:
    name: str
    url: str

    def __init__(
        self, sync_client: httpx.Client, async_client: httpx.AsyncClient, path: str
    ) -> None:
        self._sync_client = sync_client
        self._async_client = async_client
        if path.startswith("/"):
            path = path[1:]
        self.name = path.split("/")[-1]
        self.path = f"/{path}"

    def request(
        self,
        method: str,
        *,
        endpoint: str,
        params: ty.Union[None, ty.BaseModel, ty.Dict[str, ty.Any]] = None,
        data: ty.Union[None, ty.BaseModel, ty.Dict[str, ty.Any]] = None,
        files: ty.Optional[ty.Dict[str, ty.Any]] = None,
        headers: ty.Dict[str, str] = {},
        page: ty.Optional[int] = None,
        size: ty.Optional[int] = None,
        params_model: ty.Optional[ty.Type[ty.BaseModel]] = None,
        body_model: ty.Optional[ty.Type[ty.BaseModel]] = None,
        return_model: ty.Optional[ty.Type[ty.BaseModel]] = None,
        stream=False,
    ) -> ty.Union[None, ty.BaseModel, ty.Page[ty.BaseModel], ty.IO[bytes]]:
        req = self._get_validated_request(
            method=method,
            endpoint=endpoint,
            params=params,
            data=data,
            files=files,
            params_model=params_model,
            body_model=body_model,
            headers=headers,
            page=page,
            size=size,
        )
        response = self._sync_client.send(req, stream=stream)
        self._raise_and_handle_errors(response)
        content = None
        if (
            response.headers.get("content-type") == "application/json"
            and response.status_code != 204
        ):
            content = response.json()
        if return_model is not None:
            return self._convert_response_to_model(content, return_model)
        return RawIteratorIO(response.iter_bytes(chunk_size=CHUNK_SIZE))

    async def request_async(
        self,
        method: str,
        *,
        endpoint: str,
        params: ty.Union[None, ty.BaseModel, ty.Dict[str, ty.Any]] = None,
        data: ty.Union[None, ty.BaseModel, ty.Dict[str, ty.Any]] = None,
        files: ty.Optional[ty.Dict[str, ty.Any]] = None,
        headers: ty.Dict[str, str] = {},
        page: ty.Optional[int] = None,
        size: ty.Optional[int] = None,
        params_model: ty.Optional[ty.Type[ty.BaseModel]] = None,
        body_model: ty.Optional[ty.Type[ty.BaseModel]] = None,
        return_model: ty.Optional[ty.Type[ty.BaseModel]] = None,
        stream=True,
    ) -> ty.Union[None, ty.BaseModel, ty.Page[ty.BaseModel], ty.IO]:
        req = self._get_validated_request(
            method=method,
            endpoint=endpoint,
            params=params,
            data=data,
            files=files,
            params_model=params_model,
            body_model=body_model,
            headers=headers,
            page=page,
            size=size,
        )
        response = await self._async_client.send(req, stream=stream)
        self._raise_and_handle_errors(response)
        content = None
        if (
            response.headers.get("content-type") == "application/json"
            and response.status_code != 204
        ):
            content = response.json()
        if return_model is not None:
            return self._convert_response_to_model(content, return_model)
        return RawIteratorIO(response.iter_bytes(CHUNK_SIZE))

    def _get_validated_request(
        self,
        *,
        method: str,
        endpoint: str,
        params: ty.Union[None, ty.BaseModel, ty.Dict[str, ty.Any]] = None,
        data: ty.Union[None, ty.BaseModel, ty.Dict[str, ty.Any]] = None,
        files: ty.Optional[ty.Dict[str, ty.Any]] = None,
        params_model: ty.Optional[ty.Type[ty.BaseModel]] = None,
        body_model: ty.Optional[ty.Type[ty.BaseModel]] = None,
        headers: ty.Optional[ty.Dict[str, str]] = None,
        page: ty.Optional[int] = None,
        size: ty.Optional[int] = None,
    ) -> httpx.Request:
        url = self.path
        if endpoint:
            if endpoint.startswith("/"):
                url = endpoint
            else:
                url = f"{self.path}/{endpoint}"
        local_params = {}
        if page is not None:
            local_params["page"] = page
        if size is not None:
            local_params["size"] = size
        if params is not None:
            if isinstance(params, dict):
                params.update(local_params)
                if params_model is not None:
                    # if params_model is passed, use Pydantic to validate params
                    params = params_model(**params)
                    params = params.dict()
            elif isinstance(params, ty.BaseModel):
                params = params.dict(exclude_none=True, exclude_defaults=True)
                params.update(local_params)
        content = None
        if data is not None:
            if isinstance(data, dict):
                if body_model is not None:
                    body = body_model(**data)
                    content = body.json()
                else:
                    content = json.dumps(data)
            elif isinstance(data, ty.BaseModel):
                content = data.json()
        req = self._sync_client.build_request(
            method, url, headers=headers, params=params, content=content, files=files
        )
        return req

    def _raise_and_handle_errors(self, response: httpx.Response) -> None:
        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            status = e.response.status_code
            error_type = None
            sdk_error = None
            try:
                error_data = e.response.json()
            except json.JSONDecodeError:
                error_data = {}
            if isinstance(error_data, dict):
                detail = error_data.get("detail")
            else:
                detail = error_data
            if detail and isinstance(detail, dict):
                error_type = detail.get("type")
            if error_type:
                sdk_error = getattr(errors, error_type, None)
            if sdk_error:
                raise sdk_error(detail=json.dumps(detail))
            elif status == 401:
                raise errors.AuthError(detail=json.dumps(detail))
            else:
                raise e

    def _convert_response_to_model(
        self, data: ty.Any, return_model: ty.Type[ty.BaseModel]
    ) -> ty.Union[None, ty.BaseModel, ty.Page[ty.BaseModel]]:
        """Convert the JSON response to the correct Pydantic model"""
        if data is None:
            return data
        return return_model.parse_obj(data)



"""Generate a wheel for a prodigy_teams_recipes package,
including updating its meta file and requirements files."""
import json
import re
import shutil
import subprocess
import sys
import tempfile
from contextlib import contextmanager
from pathlib import Path

import pkginfo

from .. import ty


def build_wheel(src: Path, wheelhouse: Path, use_active_venv: bool) -> Path:
    with _make_tempdir() as tmp:
        if use_active_venv:
            python = sys.executable
        else:
            venv_path = tmp / "venv"
            _subprocess_run(
                [sys.executable, "-m", "venv", str(venv_path.absolute())], tmp
            )
            python = (venv_path / "bin" / "python").absolute()
            _subprocess_run(
                [str(python), "-m", "pip", "install", "--upgrade", "pip", "wheel"],
                tmp,
            )
        ret = _subprocess_run(
            [
                str(python),
                "-m",
                "pip",
                "wheel",
                str(src.absolute()),
                "-f",
                str(wheelhouse.absolute()),
                "-w",
                str(wheelhouse.absolute()),
                "--no-deps",
                "--no-cache-dir",
            ],
            tmp,
        )
    filename = re.findall(r"(?<=filename=)\S+", ret.stdout)[-1]
    wheel = wheelhouse / filename
    assert wheel.exists()
    return wheel.absolute()


def compute_recipes_meta(
    wheel: Path, wheelhouse: Path, use_active_venv: bool
) -> ty.Dict[str, ty.Any]:
    pkg = pkginfo.Wheel(str(wheel.resolve()))
    assert pkg.name is not None
    assert pkg.version is not None
    name = pkg.name
    requirements = _resolve_requirements(
        f"{name}=={pkg.version}", wheel, wheelhouse, use_active_venv
    )
    recipes_meta = _get_recipes_meta(
        name.replace("-", "_"), wheel, wheelhouse, use_active_venv
    )
    meta = {
        "name": name,
        "version": pkg.version,
        "description": pkg.summary,
        "author": pkg.author,
        "email": pkg.author_email,
        "url": pkg.home_page,
        "license": pkg.license,
        "assets": {},
        "recipes": recipes_meta,
        "requirements": requirements,
    }
    # Check it's json serializable
    _ = json.dumps(meta, indent=2)
    return meta


def _resolve_requirements(
    package_version: str, wheel: Path, wheelhouse: Path, use_active_venv: bool
) -> ty.List[str]:
    with _make_tempdir() as tmp:
        text = _exec_compile_requirements(
            sys.executable, tmp, package_version, wheel, wheelhouse, use_active_venv
        )
    lines = [line.strip() for line in text.strip().split("\n")]
    return [
        r
        for r in lines
        if r and not r.startswith("#") and not r.startswith("--find-links")
    ]


def _get_recipes_meta(
    name: str, src: Path, wheelhouse: Path, use_active_venv: bool
) -> ty.Dict[str, ty.Any]:
    with _make_tempdir() as tmp:
        _exec_get_recipes_meta(
            sys.executable,
            tmp,
            src,
            wheelhouse,
            name,
            use_active_venv,
            tmp / "recipes.json",
        )
        # Read meta
        with (tmp / "recipes.json").open("r") as file_:
            recipes = json.loads(file_.read().strip())
    return recipes


def _exec_compile_requirements(
    python: str,
    tmp: Path,
    package_version: str,
    wheel: Path,
    wheelhouse: Path,
    use_active_venv: bool,
) -> str:
    """Compile a set of requirements inside a Python virtualenv using `pip-tools`."""
    if not use_active_venv:
        venv_path = tmp / "venv"
        _subprocess_run([python, "-m", "venv", str(venv_path.absolute())], tmp)
        python = str((venv_path / "bin" / "python").absolute())
        pip_compile = str((venv_path / "bin" / "pip-compile").absolute())
        # Install pip-tools
        _subprocess_run(
            [
                python,
                "-m",
                "pip",
                "install",
                "--upgrade",
                "pip-tools",
            ],
            tmp,
        )
    else:
        pip_compile = "pip-compile"
    with (tmp / "reqs-in.txt").open("w") as file_:
        file_.write(package_version)
    # Run pip-compile
    _subprocess_run(
        [
            pip_compile,
            str((tmp / "reqs-in.txt").absolute()),
            "--output-file",
            str((tmp / "requirements.txt").absolute()),
            "-f",
            str(wheelhouse.absolute()),
            "--no-emit-find-links",
            # I was having weird problems if we don't rebuild, it seems
            # pip-tools will cache the result? That's not necessarily
            # what we want if you're developing and might not have changed
            # the version
            "--rebuild",
        ],
        tmp,
    )
    with (tmp / "requirements.txt").open(encoding="utf8") as file_:
        requirements = file_.read()
    return requirements


def _exec_get_recipes_meta(
    python: str,
    tmp: Path,
    src: Path,
    wheelhouse: Path,
    name: str,
    use_active_venv: bool,
    meta_path: Path,
) -> None:
    """Compile a set of requirements inside a Python virtualenv using `pip-tools`."""
    if not use_active_venv:
        venv_path = tmp / "venv"
        _subprocess_run([python, "-m", "venv", str(venv_path.absolute())], tmp)
        python = str((venv_path / "bin" / "python").absolute())
        # Install the package
        _subprocess_run(
            [
                python,
                "-m",
                "pip",
                "install",
                str(src.absolute()),
                "-f",
                str(wheelhouse.absolute()),
            ],
            tmp,
        )
    imports = f"{name}"
    _subprocess_run(
        [
            python,
            "-m",
            "prodigy_teams_recipes",
            "create-meta",
            imports,
            "--output",
            str(meta_path.absolute()),
        ],
        tmp,
    )


def _subprocess_run(args: ty.List[str], cwd: Path) -> subprocess.CompletedProcess:
    try:
        ret = subprocess.run(
            args,
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            cwd=cwd,
            text=True,
            encoding="utf8",
        )
    except subprocess.CalledProcessError as e:
        print(e.output)
        raise
    return ret


@contextmanager
def _make_tempdir() -> ty.Generator[Path, None, None]:
    """Execute a block in a temporary directory and remove the directory and
    its contents at the end of the with block.

    YIELDS (Path): The path of the temp directory.
    """
    d = Path(tempfile.mkdtemp())
    yield d
    shutil.rmtree(str(d))



from radicli import Arg

from ... import ty
from ...auth import get_current_auth
from ...cli import cli
from ...config import SavedSettings
from ...errors import BrokerError, CLIError
from ...messages import Messages
from ...prodigy_teams_broker_sdk import models as broker_models
from ...util import resolve_remote_path


@cli.subcommand(
    "files",
    "rm",
    remote_path=Arg(help=Messages.remote_path),
    cluster_host=Arg("--cluster-host", help=Messages.cluster_host),
    missing_ok=Arg("--missing-ok", help=Messages.missing_ok),
)
def rm(
    remote_path: str, cluster_host: ty.Optional[str] = None, missing_ok: bool = False
) -> None:
    """Remove files from the cluster"""
    broker_host = SavedSettings.get_and_save("broker_host", cluster_host)
    auth = get_current_auth()
    path = resolve_remote_path(auth.client, remote_path, broker_host)
    body = broker_models.Deleting(path=path, missing_ok=missing_ok)
    try:
        auth.broker_client.files.delete(body)
    except BrokerError as e:
        raise CLIError(Messages.E017.format(verb="delete"), e)



# flake8: noqa
# generated by datamodel-codegen:
#   filename:  openapi.json

from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional, Union
from uuid import UUID

from pydantic import BaseModel, Field


class AccessTokenResponse(BaseModel):
    access_token: str = Field(..., title="Access Token")
    token_type: Optional[str] = Field("bearer", title="Token Type")
    expires_in: int = Field(..., title="Expires In")
    expiration: Optional[int] = Field(None, title="Expiration")


class BodyApiV1UploadFileApiV1FilesUploadPost(BaseModel):
    file: bytes = Field(..., description="File to Upload with valid name", title="File")


class CheckProgressRequest(BaseModel):
    id: UUID = Field(..., title="Id")


class Status(Enum):
    running = "running"
    done = "done"


class CheckProgressResponse(BaseModel):
    id: UUID = Field(..., title="Id")
    status: Status = Field(..., title="Status")
    report: Optional[Any] = Field(None, title="Report")


class CheckStartRequest(BaseModel):
    s3_bucket: Optional[str] = Field(None, title="S3 Bucket")
    nfs_path: Optional[str] = Field(None, title="Nfs Path")
    package_environment: Optional[str] = Field(None, title="Package Environment")
    recipe_name: Optional[str] = Field(None, title="Recipe Name")
    recipe_args: Optional[Dict[str, Any]] = Field(None, title="Recipe Args")


class CheckStartResponse(BaseModel):
    id: UUID = Field(..., title="Id")


class Copying(BaseModel):
    src: str = Field(..., title="Src")
    dest: str = Field(..., title="Dest")
    make_dirs: bool = Field(..., title="Make Dirs")
    overwrite: bool = Field(..., title="Overwrite")


class Dataset(BaseModel):
    name: str = Field(
        ..., description="a unique name for the dataset", title="Dataset Name"
    )
    created: Optional[datetime] = Field(
        None,
        description="the date and time that the dataset was created",
        title="Created Time",
    )
    meta: Optional[str] = Field(
        "{}",
        description="An arbitrary JSON string containing metadata for your dataset",
        title="Dataset Metadata",
    )
    session: Optional[bool] = Field(
        False,
        description="True when the dataset contains examples for a user session",
        title="Session Dataset",
    )


class DatasetCreating(BaseModel):
    author: Optional[str] = Field(
        "", description="the user that authored created dataset", title="Dataset Author"
    )
    description: Optional[str] = Field(
        "",
        description="human readable description of what kind of examples the dataset contains",
        title="Dataset Description",
    )
    created: Optional[datetime] = Field(
        None,
        description="the date and time that the dataset was created",
        title="Created Time",
    )
    name: str = Field(
        ..., description="a unique name for the dataset", title="Dataset Name"
    )


class DatasetExample(BaseModel):
    id: int = Field(..., description="The database id", ge=1, title="Id")
    input_hash: int = Field(
        ..., description="The hash value of the input", title="Input Hash"
    )
    task_hash: int = Field(
        ..., description="The hash value of the task", title="Task Hash"
    )
    content: Dict[str, Any] = Field(
        ..., description="The dict content of the example", title="Content"
    )


class DatasetMeta(BaseModel):
    author: Optional[str] = Field(
        "", description="the user that authored created dataset", title="Dataset Author"
    )
    description: Optional[str] = Field(
        "",
        description="human readable description of what kind of examples the dataset contains",
        title="Dataset Description",
    )
    created: Optional[datetime] = Field(
        None,
        description="the date and time that the dataset was created",
        title="Created Time",
    )


class DatasetQueryResponse(BaseModel):
    items: List[DatasetExample] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")
    name: str = Field(..., title="Name")
    meta: Optional[DatasetMeta] = None


class Deleting(BaseModel):
    path: str = Field(..., title="Path")
    missing_ok: Optional[bool] = Field(True, title="Missing Ok")


class Downloading(BaseModel):
    target: str = Field(..., title="Target")


class EnvAwaiting(BaseModel):
    path: str = Field(..., title="Path")


class EnvCreating(BaseModel):
    requirements: List[str] = Field(..., title="Requirements")
    envs_path: str = Field(..., title="Envs Path")
    wheels_path: str = Field(..., title="Wheels Path")
    blocking: bool = Field(..., title="Blocking")


class EnvReturning(BaseModel):
    name: str = Field(..., title="Name")
    path: str = Field(..., title="Path")
    ready: bool = Field(..., title="Ready")


class FileCopyPlan(BaseModel):
    src: str = Field(..., title="Src")
    dest: str = Field(..., title="Dest")


class FileStats(BaseModel):
    size: int = Field(..., title="Size")
    modification_time: datetime = Field(..., title="Modification Time")


class Action(Enum):
    start = "start"
    stop = "stop"
    delete = "delete"


class JobLogRequest(BaseModel):
    id: UUID = Field(..., title="Id")


class JobLogResponse(BaseModel):
    text: str = Field(..., title="Text")


class JobType(Enum):
    task = "task"
    action = "action"


class Listing(BaseModel):
    path: str = Field(..., title="Path")
    recurse: bool = Field(..., title="Recurse")
    include_stats: Optional[bool] = Field(False, title="Include Stats")


class NomadErrorResponse(BaseModel):
    status_code: int = Field(..., title="Status Code")
    body: str = Field(..., title="Body")
    url: str = Field(..., title="Url")
    method: str = Field(..., title="Method")


class NomadJobActionResponse(BaseModel):
    action: str = Field(..., title="Action")
    job_id: UUID = Field(..., title="Job Id")
    eval_id: UUID = Field(..., title="Eval Id")
    eval_create_index: int = Field(..., title="Eval Create Index")
    job_modify_index: int = Field(..., title="Job Modify Index")


class PageStr(BaseModel):
    items: List[str] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PamErrorResponse(BaseModel):
    status_code: int = Field(..., title="Status Code")
    body: str = Field(..., title="Body")
    url: str = Field(..., title="Url")
    method: str = Field(..., title="Method")


class PathList(BaseModel):
    paths: List[str] = Field(..., title="Paths")
    stats: Optional[List[FileStats]] = Field(None, title="Stats")
    exists: bool = Field(..., title="Exists")
    is_file: bool = Field(..., title="Is File")


class RsyncPlan(BaseModel):
    copy_: List[FileCopyPlan] = Field(..., title="Copy ")
    delete: List[str] = Field(..., title="Delete")
    file_exists: List[str] = Field(..., title="File Exists")
    directory_missing: List[str] = Field(..., title="Directory Missing")


class Secret(BaseModel):
    key: str = Field(..., title="Key")
    value: str = Field(..., title="Value")


class SecretResponse(BaseModel):
    CreateIndex: int = Field(..., title="Createindex")
    ModifyIndex: int = Field(..., title="Modifyindex")
    LockIndex: int = Field(..., title="Lockindex")
    Key: str = Field(..., title="Key")
    Flags: int = Field(..., title="Flags")
    Value: str = Field(..., title="Value")


class SpecValidationFailure(BaseModel):
    message: str = Field(..., title="Message")
    errors: List[Dict[str, Any]] = Field(..., title="Errors")


class Statting(BaseModel):
    path: str = Field(..., title="Path")


class StatusModel(Enum):
    Ready = "Ready"
    Error = "Error"


class TasksBaseUrl(BaseModel):
    url: str = Field(..., title="Url")


class ValidationError(BaseModel):
    loc: List[Union[str, int]] = Field(..., title="Location")
    msg: str = Field(..., title="Message")
    type: str = Field(..., title="Error Type")


class CopyPlan(BaseModel):
    copy_: List[FileCopyPlan] = Field(..., title="Copy ")
    file_exists: List[str] = Field(..., title="File Exists")
    directory_missing: List[str] = Field(..., title="Directory Missing")


class HTTPValidationError(BaseModel):
    detail: Optional[List[ValidationError]] = Field(None, title="Detail")


class HealthCheck(BaseModel):
    status: StatusModel
    cluster: StatusModel
    nomad_error: Optional[NomadErrorResponse] = None


class JobActionResponse(BaseModel):
    job_id: UUID = Field(..., title="Job Id")
    attempt_id: Optional[UUID] = Field(None, title="Attempt Id")
    status: str = Field(..., title="Status")
    action: Action = Field(..., title="Action")
    job_type: JobType
    pam_error: Optional[PamErrorResponse] = None
    validation_error: Optional[SpecValidationFailure] = None
    nomad_error: Optional[NomadErrorResponse] = None
    nomad_change: Optional[NomadJobActionResponse] = None


class JobID(BaseModel):
    id: UUID = Field(..., title="Id")
    job_type: JobType



from .. import ty
from ..models import (
    ActionReturning,
    ClusterJobReading,
    EmptyResponse,
    NomadAttemptUpdating,
    NomadEventBatchCreating,
    NomadEventGetIndex,
    NomadEventIndex,
    NomadJobs,
    ReportNomadJobResponse,
    TaskReturning,
)
from .base import BaseClient


class ForCluster(BaseClient):
    class Token(ty.BaseModel):
        token: str

    def read_task(self, data: ClusterJobReading, **kwargs: ty.Any) -> TaskReturning:
        res = self.request(
            "POST",
            endpoint="read-task",
            data=data,
            return_model=TaskReturning,
        )
        return ty.cast(TaskReturning, res)

    async def read_task_async(
        self, data: ClusterJobReading, **kwargs: ty.Any
    ) -> TaskReturning:
        res = await self.request_async(
            "POST",
            endpoint="read-task",
            data=data,
            return_model=TaskReturning,
        )
        return ty.cast(TaskReturning, res)

    def read_action(self, data: ClusterJobReading, **kwargs: ty.Any) -> ActionReturning:
        res = self.request(
            "POST",
            endpoint="read-action",
            data=data,
            return_model=ActionReturning,
        )
        return ty.cast(ActionReturning, res)

    async def read_action_async(
        self, data: ClusterJobReading, **kwargs: ty.Any
    ) -> ActionReturning:
        res = await self.request_async(
            "POST",
            endpoint="read-action",
            data=data,
            return_model=TaskReturning,
        )
        return ty.cast(ActionReturning, res)

    def report_nomad_event(
        self, data: NomadEventBatchCreating, **kwargs: ty.Any
    ) -> None:
        self.request(
            "POST",
            endpoint="report-nomad-event",
            data=data,
        )
        return None

    async def report_nomad_event_async(
        self, data: NomadEventBatchCreating, **kwargs: ty.Any
    ) -> None:
        await self.request_async(
            "POST",
            endpoint="report-nomad-event",
            data=data,
        )
        return None

    def update_attempt(
        self, data: NomadAttemptUpdating, **kwargs: ty.Any
    ) -> EmptyResponse:
        res = self.request(
            "POST",
            endpoint="update-attempt",
            data=data,
            return_model=EmptyResponse,
        )
        return ty.cast(EmptyResponse, res)

    async def update_attempt_async(
        self, data: NomadAttemptUpdating, **kwargs: ty.Any
    ) -> EmptyResponse:
        res = await self.request_async(
            "POST",
            endpoint="update-attempt",
            data=data,
            return_model=EmptyResponse,
        )
        return ty.cast(EmptyResponse, res)

    def report_nomad_jobs(
        self, data: ty.Union[NomadJobs, dict], **kwargs: ty.Any
    ) -> ReportNomadJobResponse:
        res = self.request(
            "POST",
            endpoint="report-nomad-jobs",
            data=data,
        )
        return ty.cast(ReportNomadJobResponse, res)

    async def report_nomad_jobs_async(
        self, data: ty.Union[NomadJobs, dict], **kwargs: ty.Any
    ) -> ReportNomadJobResponse:
        res = await self.request_async(
            "POST",
            endpoint="report-nomad-jobs",
            data=data,
            return_model=ReportNomadJobResponse,
        )
        return ty.cast(ReportNomadJobResponse, res)

    def get_nomad_event_index(self, data: NomadEventGetIndex) -> NomadEventIndex:
        res = self.request(
            "POST",
            endpoint="get-nomad-event-index",
            data=data,
            return_model=NomadEventIndex,
        )
        return ty.cast(NomadEventIndex, res)

    async def get_nomad_event_index_async(
        self, data: NomadEventGetIndex
    ) -> NomadEventIndex:
        res = await self.request_async(
            "POST",
            endpoint="get-nomad-event-index",
            data=data,
            return_model=NomadEventIndex,
        )
        return ty.cast(NomadEventIndex, res)



import inspect
import re
from enum import Enum

from pydantic import validator
from radicli import (
    DEFAULT_PLACEHOLDER,
    Arg,
    ArgparseArg,
    format_type,
    get_arg,
    get_list_converter,
)
from radicli.util import BASE_TYPES, join_strings

from . import models, ty
from .errors import RecipeProcessingError

SIMPLE_TYPES = {str: "str", int: "int", bool: "bool", float: "float", ty.UUID: "uuid"}
SIMPLE_TYPES_BY_NAME = {v: k for k, v in SIMPLE_TYPES.items()}
OBJECT_FIELD_KEYS = ("asset", "dataset", "secret")
UNPARSED = object()

# We can pass in Dict values because we might have Asset, Dataset
# etc fields in there.
PlanType = ty.Dict[str, ty.Union[ty.Dict, str, int, float, bool, ty.List[str]]]
ObjectsMapType = ty.Dict[str, ty.Dict[str, ty.Dict[str, ty.Any]]]
TypeCallbackType = ty.Callable[[str], ty.Optional[ty.Callable[..., ty.Any]]]
_InT = ty.TypeVar("_InT", bound=ty.Union[str, int, float])
_CustomT = ty.TypeVar("_CustomT")


def slugify_name(v: str) -> str:
    """Turn a user input readable string into a URL friendly slug without invalid characters"""
    input_name = re.sub("_", "-", v)
    input_name = re.sub(r"[^A-Za-z0-9\-\. ]", "", input_name)
    input_name = re.sub(r"\s+", "-", input_name)
    input_name = input_name.lower()
    return input_name


class Validation(models.ValidationSchema):
    # fmt: off
    op: ty.Union[models.ValidationOp, ty.Literal["ge", "gt", "le", "lt", "eq", "ne", "re"]]
    level: ty.Union[models.ValidationLevel, ty.Literal["error", "warning", "info", "success"]]
    # fmt: on


# Frustratingly, the SDK conversion process messes up defaults, which
# hurts quite a bit for these. So we redefine and set defaults.


class Props(ty.BaseModel):
    """Properties common to all types"""

    # Set defaults for type checker
    name: ty.Optional[str] = None
    title: ty.Optional[str] = None
    description: ty.Optional[str] = None
    placeholder: ty.Optional[str] = None
    value: ty.Optional[ty.Any] = None
    required: ty.Optional[bool] = None
    positional: ty.Optional[bool] = None
    exists: ty.Optional[bool] = None
    validations: ty.Optional[ty.List[models.ValidationSchema]] = None
    pattern: ty.Optional[str] = None  # HTML default regex validation
    widget: ty.Optional[str] = None
    choice: ty.Optional[ty.Dict[str, "Props"]] = None
    min: ty.Optional[float] = None
    max: ty.Optional[float] = None
    step: ty.Optional[float] = None
    optional_title: ty.Optional[str] = None


class OptionalProps(Props):
    """Properties for optional fields that provide an additional label for the toggle"""

    optional_title: ty.Optional[str] = None


class TextProps(Props):
    """Properties for text fields"""

    pattern: ty.Optional[str] = None  # HTML default regex validation
    widget: ty.Optional[ty.Literal["text", "textarea", "password"]] = None


class IntProps(Props):
    """Properties for int fields"""

    step: int = 1
    min: ty.Optional[int] = None
    max: ty.Optional[int] = None
    widget: ty.Optional[ty.Literal["number", "range"]] = None


class FloatProps(Props):
    """Properties for float fields"""

    step: float = 0.01
    min: ty.Optional[float] = None
    max: ty.Optional[float] = None
    widget: ty.Optional[ty.Literal["number", "range"]] = None


class BoolProps(Props):
    """Properties for bool fields"""

    widget: ty.Optional[ty.Literal["checkbox", "toggle"]] = None


class ChoiceProps(Props):
    """Properties for choice fields like Enum and Union"""

    # Additional properties for each choice
    choice: ty.Optional[ty.Dict[str, "AnyProps"]] = None
    widget: ty.Optional[ty.Literal["select", "radio"]] = None


class ListProps(Props):
    """Properties for List fields"""

    min: ty.Optional[int] = None  # Minimum number of items in the list
    max: ty.Optional[int] = None  # Maximum number of items in the list


AnyProps = ty.Union[
    IntProps, FloatProps, BoolProps, TextProps, ChoiceProps, ListProps, Props
]
ChoiceProps.update_forward_refs()


class Field(models.FieldSchema):
    # Set defaults for type checker compatibility
    props: Props
    fields: ty.List["Field"] = []
    choice: ty.List["Field"] = []

    # Subclassing the field here so we can slugify names
    @validator("name")
    def slugify_name(cls, v: str) -> str:
        return slugify_name(v)


class RecipeSchema(models.RecipeSchema):
    fields: ty.List[Field]
    title: ty.Optional[str] = None
    description: ty.Optional[str] = None
    view_id: ty.Optional[str] = None
    cli_names: ty.Dict[str, str] = {}


def get_field_optional(schema: models.RecipeSchema) -> ty.Dict[str, bool]:
    """Walk the schema, and map field names according to whether they're globally optional.
    A field is globally optional if it descends from a choice or a non-required
    field.
    """
    output = {}
    queue = [(not field.props.required, field) for field in schema.fields]
    for is_optional, field in queue:
        output[field.name] = is_optional
        # Fields of a custom type are optional if they descend from an optional, or
        # if they're not required
        sub_fields = field.fields or []
        queue.extend([(is_optional or not f.props.required, f) for f in sub_fields])
        # Everything descending from a choice (e.g. an object nested within a union) needs to
        # be marked optional
        queue.extend([(True, f) for f in (field.choice or [])])
    return output


def get_field_names(field: Field) -> ty.Set[str]:
    """Walk the field schema and get all of its parent and child names."""
    output = set()
    queue = [(f.name, f) for f in field.fields] + [(f.name, f) for f in field.choice]
    for name, field in queue:
        output.add(name)
        queue.extend(
            [(f.name, f) for f in field.fields] + [(f.name, f) for f in field.choice]
        )
    return output


def get_field_cli_type(
    field: Field, parent: ty.Optional[Field] = None
) -> ty.Tuple[ty.Type, ty.Optional[ty.Callable[[str], ty.Any]]]:
    if field.type.name in SIMPLE_TYPES_BY_NAME:
        return SIMPLE_TYPES_BY_NAME[field.type.name], None
    elif field.type.object_type in OBJECT_FIELD_KEYS:
        return str, None
    elif field.type.name == "list":
        type_args = ty.cast(ty.List[models.TypeSchema], field.type.args)
        arg_name = type_args[0].name if len(type_args) else "str"
        base_type = SIMPLE_TYPES_BY_NAME.get(arg_name, str)
        return ty.List[base_type], get_list_converter(base_type)
    elif field.type.name == "enum":
        enum_args = {f.name: f.props.value for f in field.choice}
        enum = Enum("str", enum_args)  # name for CLI help display purpose only
        # We currently expect all enum members to have the same type
        enum_type = type(field.choice[0].props.value) if field.choice else str
        base_type = enum_type if enum_type and enum_type in SIMPLE_TYPES else str
        return enum, lambda v: base_type(getattr(enum, v).value)
    elif field.type.name == "literal" and field.type.args:
        # Only handling string literals (also supported in schema only)
        enum = Enum("str", {k: k for k in field.type.args})
        return enum, lambda v: getattr(enum, v).value
    elif parent is not None and parent.type.name in ["union", "optional"]:
        # If we're an empty type within a union, make it a bool
        return bool, None
    else:
        return str, None


def get_field_cli_args(
    field: Field,
    cli_names: ty.Dict[str, str],
    name: ty.Optional[str] = None,
    parent: ty.Optional[Field] = None,
    force_optional: bool = False,
) -> ty.List[ArgparseArg]:
    name = name or field.props.name or field.name
    # If a field is optional, force everything underneath it to be optional
    force_optional = force_optional or not field.props.required
    if field.type.object_type in OBJECT_FIELD_KEYS:
        # Handle these as special cases: we only ask for their name
        return [
            _create_field_arg(
                cli_names.get(name, name), str, None, field.props, force_optional
            )
        ]
    elif len(field.fields) > 0:
        args = []
        # If the custom class is Optional we add the option of creating it with defaults
        if parent is not None and parent.type.name == "optional":
            arg_type, arg_converter = get_field_cli_type(field, parent=parent)
            if name in cli_names:
                name = cli_names[name]
            args.append(
                _create_field_arg(
                    name, arg_type, arg_converter, field.props, force_optional
                )
            )
        for subfield in field.fields:
            args.extend(
                get_field_cli_args(
                    subfield, cli_names, parent=field, force_optional=force_optional
                )
            )
        return args
    elif hasattr(field, "choice") and field.choice and field.type.name != "enum":
        args = []
        for choice_field in field.choice:
            # If we have a union or enum, everything deriving from it is optional
            choice_args = get_field_cli_args(
                choice_field, cli_names, parent=field, force_optional=True
            )
            args.extend(choice_args)
        return args
    else:
        arg_type, arg_converter = get_field_cli_type(field, parent=parent)
        if name in cli_names:
            name = cli_names[name]
        return [
            _create_field_arg(
                name, arg_type, arg_converter, field.props, force_optional
            )
        ]


def _create_field_arg(
    name: str, arg_type, arg_converter, props: AnyProps, force_optional: bool
) -> ArgparseArg:
    if arg_type == ty.UUID:  # TODO: Shouldn't we allow UUID?
        arg_type = str
    help_texts = [props.title, props.description]
    arg_help = ": ".join([t for t in help_texts if t is not None])
    arg_opt = f"--{name}" if not props.positional or arg_type is bool else None
    arg = Arg(arg_opt, help=arg_help if arg_help else None, converter=arg_converter)
    default = (
        props.value
        if (force_optional or props.value is not None or not props.required)
        else DEFAULT_PLACEHOLDER
    )
    has_converter = arg_converter is not None and arg_converter not in BASE_TYPES
    ap_arg = get_arg(
        name,
        arg,
        arg_converter if has_converter else arg_type,
        orig_type=arg_type,
        default=default,
        has_converter=has_converter,
        skip_resolve=has_converter,
    )
    ap_arg.help = join_strings(
        arg.help,
        f"({format_type(arg_type)})",
        f"(default: {default})" if default != DEFAULT_PLACEHOLDER else "",
    )
    return ap_arg


def parse_plan(
    schema: RecipeSchema,
    plan: PlanType,
    objects_map: ObjectsMapType,
    get_custom_type: TypeCallbackType,
) -> ty.Tuple[ty.List[str], ty.Dict]:
    # This doesn't deal with defaults or validation at all – we assume that
    # those aspects are handled by the function itself.
    plan = dict(plan)
    args = []
    kwargs: ty.Dict[str, ty.Any] = {}
    plan = resolve_cli_names(plan, schema.cli_names)
    for field in schema.fields:
        value = _parse_field(field, plan, objects_map, get_custom_type)
        if value is UNPARSED and not field.props.required:
            continue
        if value is UNPARSED:  # TODO: improve errors
            raise RecipeProcessingError(f"Could not parse field '{field.name}'", plan)
        if isinstance(value, ty.UUID):
            value = str(value)
        if field.props.positional:
            args.append(value)
        else:
            kwargs[field.name.replace("-", "_")] = value
    return args, kwargs


def _parse_field(
    field: Field,
    plan: PlanType,
    objects_map: ObjectsMapType,
    get_custom_type: TypeCallbackType,
) -> ty.Any:
    """Parse a given schema field using the argument values provided."""
    object_type = field.type.object_type
    if field.type.name == "union":  # try parsing all fields of a union
        maybe_values = []
        for choice in field.choice:
            maybe_value = _parse_field(choice, plan, objects_map, get_custom_type)
            if maybe_value is not UNPARSED:
                maybe_values.append((choice, maybe_value))
        if not maybe_values:  # no union fields could be parsed
            return UNPARSED
        elif len(maybe_values) == 1:  # only one field could be parsed
            return maybe_values[0][1]
        else:  # figure out which field to return
            for maybe_field, maybe_value in maybe_values:
                for field_name in get_field_names(maybe_field):
                    if field_name in plan:
                        return maybe_value
                if maybe_field.name in plan:  # the union choice
                    return maybe_value
    if field.type.name == "optional":  # if value is provided, parse as original type
        maybe_field = field.choice[0]
        maybe_value = _parse_field(maybe_field, plan, objects_map, get_custom_type)
        for field_name in get_field_names(maybe_field):
            if field_name in plan:
                return maybe_value
        if maybe_field.name in plan:
            return maybe_value
        return UNPARSED
    elif object_type is not None and field.name in plan:
        name = plan[field.name]
        if isinstance(name, ty.UUID):
            name = str(name)
        if not isinstance(name, str):
            err = f"Invalid object field '{name}' ({type(name)})"
            raise RecipeProcessingError(err, plan)
        object_data = objects_map.get(object_type, {}).get(name)
        if object_data is None:
            err = f"Can't find data for {object_type} '{name}' in objects map"
            err_info = (
                f"This can happen when running a recipe without a connection to "
                f"the cluster. You can provide a path to a JSON file keyed by "
                f"object type with data to use for named objects via the "
                f"PRODIGY_TEAMS_RECIPES_OBJECTS_PATH environment variable. For example:\n"
                f'{{"{object_type}": {{"{name}": {{"name": "{name}", ...}}}}}}'
            )
            raise RecipeProcessingError(err, err_info)
        assert isinstance(object_data, dict)
        custom_type = get_custom_type(field.type.name)
        keys = {"name", "id", "broker_id", "version", "path", "meta"}
        value = {k: v for k, v in object_data.items() if k in keys}
        return call_custom_type(custom_type, value)
    elif object_type is not None:
        return UNPARSED
    elif field.fields:  # parse subfields
        value = {}
        for subfield in field.fields:
            maybe_value = _parse_field(subfield, plan, objects_map, get_custom_type)
            if maybe_value is UNPARSED and subfield.props.required:
                return UNPARSED
            elif maybe_value is not UNPARSED:
                value[subfield.name.rsplit(".", 1)[1]] = maybe_value
        custom_type = get_custom_type(field.type.name)
        return call_custom_type(custom_type, value)
    elif field.type.name == "list":
        return _parse_list_field(field, plan, objects_map, get_custom_type)
    elif field.name in plan:
        value = plan[field.name]
        custom_type = get_custom_type(field.type.name)
        if custom_type is not None and inspect.isclass(
            custom_type
        ):  # the custom type doesn't have fields
            if value is True:
                # we can receive the choice as bool but it's actually an object
                value = {}
            assert isinstance(value, dict)
            return call_custom_type(custom_type, value)
        return value
    else:
        return UNPARSED


def _parse_list_field(
    field: Field,
    plan: PlanType,
    objects_map: ObjectsMapType,
    get_custom_type: TypeCallbackType,
) -> ty.Any:
    if field.name not in plan:
        return UNPARSED
    list_values = plan[field.name]
    if not isinstance(list_values, list):
        err = f"Received invalid list value for '{field.name}'"
        raise RecipeProcessingError(err, list_values)
    if field.type.args and isinstance(field.type.args[0], models.TypeSchema):
        result = []
        for item in list_values:
            dummy = Field(name=field.name, type=field.type.args[0], props=field.props)
            value = _parse_field(
                dummy, {field.name: item}, objects_map, get_custom_type
            )
            if value is UNPARSED:
                return UNPARSED
            result.append(value)
        return result
    return list_values


def call_custom_type(
    custom_type: ty.Optional[ty.Callable[..., _CustomT]], values: ty.Dict[str, ty.Any]
) -> _CustomT:
    assert custom_type is not None
    values = {key.replace("-", "_"): value for key, value in values.items()}
    try:
        return custom_type(**values)
    except Exception as e:
        raise RecipeProcessingError(
            f"Failed to call custom type {custom_type} with values {values}. Error: {e}"
        ) from e


def resolve_cli_names(
    args: ty.Dict[str, ty.Any], cli_names: ty.Optional[ty.Dict[str, str]]
) -> ty.Dict[str, ty.Any]:
    if not cli_names:
        return args
    inverse_cli_names = {v: k for k, v in cli_names.items()}
    new_args = {}
    for key, value in args.items():
        new_args[inverse_cli_names.get(key, key)] = value
    return new_args



# flake8: noqa
# generated by datamodel-codegen:
#   filename:  openapi.json

from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional, Union
from uuid import UUID

from pydantic import BaseModel, EmailStr, Field


class ActionDeleting(BaseModel):
    id: UUID = Field(..., title="Id")


class ActionState(Enum):
    NOT_FOUND = "NOT_FOUND"
    CREATED = "CREATED"
    PREPARING = "PREPARING"
    WITH_PROBLEMS = "WITH_PROBLEMS"
    STARTING = "STARTING"
    STARTED = "STARTED"
    NOT_STARTED = "NOT_STARTED"
    STOPPING = "STOPPING"
    STOPPED = "STOPPED"
    DELETING = "DELETING"
    PENDING = "PENDING"
    FAILED = "FAILED"
    NOMAD_ERROR = "NOMAD_ERROR"
    NOMAD_JOB_ERROR = "NOMAD_JOB_ERROR"
    OBTAINING_MODELS_VENV = "OBTAINING_MODELS_VENV"
    INITIALIZING_VENV = "INITIALIZING_VENV"
    COMPLETE = "COMPLETE"
    CANCELLED = "CANCELLED"
    EXECUTION_NOT_FOUND = "EXECUTION_NOT_FOUND"
    EXECUTION_RESULT_ERROR = "EXECUTION_RESULT_ERROR"
    NO_RESULTS_PATH_VALID = "NO_RESULTS_PATH_VALID"
    NO_EXECUTION = "NO_EXECUTION"


class ActionStateReturning(BaseModel):
    updated: datetime = Field(..., title="Updated")
    id: UUID = Field(..., title="Id")
    state: ActionState
    is_running: bool = Field(..., title="Is Running")
    is_startable: bool = Field(..., title="Is Startable")
    is_stoppable: bool = Field(..., title="Is Stoppable")


class ActionStats(BaseModel):
    executions: int = Field(..., title="Executions")
    last_run: Optional[datetime] = Field(None, title="Last Run")


class ActionUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    name: Optional[str] = Field(None, title="Name")


class AllocClientStatus(Enum):
    pending = "pending"
    running = "running"
    complete = "complete"
    failed = "failed"
    lost = "lost"
    unknown = "unknown"


class AllocDesiredStatus(Enum):
    run = "run"
    stop = "stop"
    evict = "evict"


class AssetCreating(BaseModel):
    broker_id: UUID = Field(..., title="Broker Id")
    name: str = Field(..., title="Name")
    version: str = Field(..., title="Version")
    path: str = Field(..., title="Path")
    kind: str = Field(..., title="Kind")
    meta: Dict[str, Any] = Field(..., title="Meta")


class AssetDeleting(BaseModel):
    id: UUID = Field(..., title="Id")


class AssetListingLatest(BaseModel):
    org_id: Optional[UUID] = Field(None, title="Org Id")
    broker_id: Optional[UUID] = Field(None, title="Broker Id")


class AssetReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    broker_id: Optional[UUID] = Field(None, title="Broker Id")
    name: Optional[str] = Field(None, title="Name")
    version: Optional[str] = Field(None, title="Version")
    kind: Optional[str] = Field(None, title="Kind")
    path: Optional[str] = Field(None, title="Path")


class AssetReadingLatest(BaseModel):
    name: str = Field(..., title="Name")
    broker_id: UUID = Field(..., title="Broker Id")


class AssetReturning(BaseModel):
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    id: UUID = Field(..., title="Id")
    broker_id: UUID = Field(..., title="Broker Id")
    name: str = Field(..., title="Name")
    version: str = Field(..., title="Version")
    kind: str = Field(..., title="Kind")
    path: str = Field(..., title="Path")
    meta: Dict[str, Any] = Field(..., title="Meta")
    num_used_by: int = Field(..., title="Num Used By")


class AssetUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    name: Optional[str] = Field(None, title="Name")
    version: Optional[str] = Field(None, title="Version")
    kind: Optional[str] = Field(None, title="Kind")
    path: Optional[str] = Field(None, title="Path")
    meta: Optional[Dict[str, Any]] = Field(None, title="Meta")


class BodyApiV1OauthTokenApiV1OauthTokenPost(BaseModel):
    grant_type: Optional[str] = Field(
        "client_credentials",
        description="OAuth2 Grant Type",
        regex="client_credentials",
        title="Grant Type",
    )
    scope: Optional[str] = Field("", description="OAuth2 Scope", title="Scope")
    client_id: Optional[str] = Field(
        None, description="Prodigy Teams Broker Client ID", title="Client Id"
    )
    client_secret: Optional[str] = Field(
        None, description="Prodigy Teams Broker Client Secret", title="Client Secret"
    )


class BrokerAuthenticated(BaseModel):
    access_token: str = Field(..., title="Access Token")
    expires_in: int = Field(..., title="Expires In")
    token_type: Optional[str] = Field("bearer", title="Token Type")


class BrokerDeleting(BaseModel):
    id: UUID = Field(..., title="Id")


class BrokerPathCreating(BaseModel):
    broker_id: UUID = Field(..., title="Broker Id")
    name: str = Field(..., title="Name")
    path: str = Field(..., title="Path")


class BrokerPathDeleting(BaseModel):
    id: UUID = Field(..., title="Id")


class BrokerPathReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    broker_id: Optional[UUID] = Field(None, title="Broker Id")
    name: Optional[str] = Field(None, title="Name")
    path: Optional[str] = Field(None, title="Path")


class BrokerPathReturning(BaseModel):
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    id: UUID = Field(..., title="Id")
    broker_id: UUID = Field(..., title="Broker Id")
    name: str = Field(..., title="Name")
    path: str = Field(..., title="Path")


class BrokerPathUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    name: str = Field(..., title="Name")
    path: str = Field(..., title="Path")


class BrokerReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    org_id: Optional[UUID] = Field(None, title="Org Id")
    address: Optional[str] = Field(None, title="Address")
    state: Optional[str] = Field(None, title="State")
    region: Optional[str] = Field(None, title="Region")
    zone: Optional[str] = Field(None, title="Zone")
    cloud_provider: Optional[str] = Field(None, title="Cloud Provider")
    cloud_account: Optional[str] = Field(None, title="Cloud Account")
    cloud_region: Optional[str] = Field(None, title="Cloud Region")


class BrokerRegistering(BaseModel):
    address: str = Field(..., title="Address")
    state: str = Field(..., title="State")


class BrokerRegistrationCreating(BaseModel):
    name: str = Field(..., title="Name")
    cloud_provider: str = Field(..., title="Cloud Provider")
    cloud_account: Optional[str] = Field(None, title="Cloud Account")
    cloud_region: Optional[str] = Field(None, title="Cloud Region")


class BrokerRegistrationReturning(BaseModel):
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    org_id: UUID = Field(..., title="Org Id")
    name: str = Field(..., title="Name")
    registration_token: UUID = Field(..., title="Registration Token")
    claimed: bool = Field(..., title="Claimed")
    expiration: datetime = Field(..., title="Expiration")


class BrokerReturning(BaseModel):
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    id: UUID = Field(..., title="Id")
    org_id: UUID = Field(..., title="Org Id")
    name: str = Field(..., title="Name")
    address: str = Field(..., title="Address")
    state: str = Field(..., title="State")
    cloud_provider: str = Field(..., title="Cloud Provider")
    cloud_account: Optional[str] = Field(None, title="Cloud Account")
    cloud_region: Optional[str] = Field(None, title="Cloud Region")
    client_id: UUID = Field(..., title="Client Id")
    client_secret: str = Field(..., title="Client Secret")


class BrokerUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    name: Optional[str] = Field(None, title="Name")
    address: Optional[str] = Field(None, title="Address")


class CackQuerying(BaseModel):
    source: str = Field(..., title="Source")
    required: bool = Field(..., title="Required")
    fields: Optional[List[str]] = Field(None, title="Fields")
    filters: Optional[Dict[str, Any]] = Field({}, title="Filters")
    predicates: Optional[List[Dict[str, Any]]] = Field([], title="Predicates")
    relations: Optional[List[CackQuerying]] = Field([], title="Relations")
    key: Optional[str] = Field("", title="Key")
    return_list: Optional[bool] = Field(None, title="Return List")
    slice_start: Optional[int] = Field(None, title="Slice Start")
    slice_stop: Optional[int] = Field(None, title="Slice Stop")
    slice_step: Optional[int] = Field(None, title="Slice Step")


class DatasetCreating(BaseModel):
    name: str = Field(..., title="Name")
    broker_id: UUID = Field(..., title="Broker Id")
    kind: str = Field(..., title="Kind")


class DatasetDeleting(BaseModel):
    id: UUID = Field(..., title="Id")


class DatasetReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    name: Optional[str] = Field(None, title="Name")
    kind: Optional[str] = Field(None, title="Kind")
    broker_id: Optional[UUID] = Field(None, title="Broker Id")


class DatasetReturning(BaseModel):
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    id: UUID = Field(..., title="Id")
    name: str = Field(..., title="Name")
    broker_id: UUID = Field(..., title="Broker Id")
    kind: str = Field(..., title="Kind")
    num_used_by: int = Field(..., title="Num Used By")


class DatasetUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    name: Optional[str] = Field(None, title="Name")
    kind: Optional[str] = Field(None, title="Kind")


class Empty(BaseModel):
    pass


class EmptyResponse(BaseModel):
    pass


class GroupAuthorizing(BaseModel):
    id: UUID = Field(..., title="Id")
    role: str = Field(..., title="Role")
    resource_type: str = Field(..., title="Resource Type")
    resource_id: UUID = Field(..., title="Resource Id")


class GroupCreating(BaseModel):
    name: str = Field(..., title="Name")
    org_id: UUID = Field(..., title="Org Id")


class GroupDeleting(BaseModel):
    id: UUID = Field(..., title="Id")


class GroupPopulating(BaseModel):
    id: UUID = Field(..., title="Id")
    user_id: UUID = Field(..., title="User Id")


class GroupReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    name: Optional[str] = Field(None, title="Name")
    org_id: Optional[UUID] = Field(None, title="Org Id")


class GroupUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    name: Optional[str] = Field(None, title="Name")


class InvitationCreating(BaseModel):
    org_id: UUID = Field(..., title="Org Id")
    email: str = Field(..., title="Email")
    hash: str = Field(..., title="Hash")
    strict: Optional[bool] = Field(False, title="Strict")
    vars: Optional[Dict[str, Any]] = Field({}, title="Vars")


class InvitationDeleting(BaseModel):
    id: UUID = Field(..., title="Id")


class InvitationReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    org_id: Optional[UUID] = Field(None, title="Org Id")
    email: Optional[str] = Field(None, title="Email")
    hash: Optional[str] = Field(None, title="Hash")
    claimed: Optional[bool] = Field(None, title="Claimed")


class InvitationReturning(BaseModel):
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    id: UUID = Field(..., title="Id")
    org_id: UUID = Field(..., title="Org Id")
    email: str = Field(..., title="Email")
    hash: str = Field(..., title="Hash")
    claimed: bool = Field(..., title="Claimed")


class InvitationUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    claimed: bool = Field(..., title="Claimed")


class JobStatus(Enum):
    pending = "pending"
    running = "running"
    dead = "dead"


class JobType(Enum):
    task = "task"
    action = "action"


class NomadAttemptResultReading(BaseModel):
    id: UUID = Field(..., title="Id")
    eval_id: UUID = Field(..., title="Eval Id")


class NomadAttemptResultReturning(BaseModel):
    success: bool = Field(..., title="Success")
    error_message: Optional[str] = Field(None, title="Error Message")


class NomadAttemptUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    pam_error: Optional[Dict[str, Any]] = Field(None, title="Pam Error")
    validation_error: Optional[Dict[str, Any]] = Field(None, title="Validation Error")
    nomad_error: Optional[Dict[str, Any]] = Field(None, title="Nomad Error")
    evaluation_id: Optional[UUID] = Field(None, title="Evaluation Id")
    eval_create_index: Optional[int] = Field(None, title="Eval Create Index")
    job_modify_index: Optional[int] = Field(None, title="Job Modify Index")


class NomadEventCreating(BaseModel):
    broker_id: UUID = Field(..., title="Broker Id")
    index: int = Field(..., title="Index")
    topic: str = Field(..., title="Topic")
    type: str = Field(..., title="Type")
    key: str = Field(..., title="Key")
    payload: Dict[str, Any] = Field(..., title="Payload")
    filter_keys: Optional[List[str]] = Field([], title="Filter Keys")
    namespace: Optional[str] = Field("", title="Namespace")


class NomadEventDetail(BaseModel):
    created: Optional[datetime] = Field(None, title="Created")
    updated: Optional[datetime] = Field(None, title="Updated")
    id: Optional[UUID] = Field(None, title="Id")
    broker_id: Optional[UUID] = Field(None, title="Broker Id")
    action_id: Optional[UUID] = Field(None, title="Action Id")
    task_id: Optional[UUID] = Field(None, title="Task Id")
    index: Optional[int] = Field(None, title="Index")
    topic: Optional[str] = Field(None, title="Topic")
    type: Optional[str] = Field(None, title="Type")
    key: Optional[str] = Field(None, title="Key")
    namespace: Optional[str] = Field(None, title="Namespace")
    payload: Dict[str, Any] = Field(..., title="Payload")


class NomadEventGetIndex(BaseModel):
    latest_index: int = Field(..., title="Latest Index")


class NomadEventIndex(BaseModel):
    index: int = Field(..., title="Index")


class NomadEventReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    broker_id: Optional[UUID] = Field(None, title="Broker Id")
    action_id: Optional[UUID] = Field(None, title="Action Id")
    task_id: Optional[UUID] = Field(None, title="Task Id")
    index: Optional[int] = Field(None, title="Index")
    topic: Optional[str] = Field(None, title="Topic")
    type: Optional[str] = Field(None, title="Type")
    key: Optional[str] = Field(None, title="Key")
    namespace: Optional[str] = Field(None, title="Namespace")


class NomadJobStub(BaseModel):
    CreateIndex: int = Field(..., title="Createindex")
    Datacenters: List[str] = Field(..., title="Datacenters")
    ID: UUID = Field(..., title="Id")
    JobModifyIndex: int = Field(..., title="Jobmodifyindex")
    ModifyIndex: int = Field(..., title="Modifyindex")
    Name: str = Field(..., title="Name")
    Namespace: Optional[str] = Field(None, title="Namespace")
    ParentID: Optional[UUID] = Field(None, title="Parentid")
    Periodic: Optional[bool] = Field(None, title="Periodic")
    Priority: Optional[int] = Field(None, title="Priority")
    Stable: Optional[bool] = Field(None, title="Stable")
    Status: JobStatus
    StatusDescription: str = Field(..., title="Statusdescription")
    Stop: bool = Field(..., title="Stop")
    SubmitTime: int = Field(..., title="Submittime")
    Type: Optional[str] = Field(None, title="Type")


class NomadJobsIds(BaseModel):
    jobs_ids: List[UUID] = Field(..., title="Jobs Ids")


class NotificationCreating(BaseModel):
    user_id: UUID = Field(..., title="User Id")
    type: str = Field(..., title="Type")
    message: str = Field(..., title="Message")
    is_read: bool = Field(..., title="Is Read")
    details: Dict[str, Union[str, int, float]] = Field(..., title="Details")


class NotificationDeleting(BaseModel):
    id: UUID = Field(..., title="Id")


class NotificationReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    user_id: Optional[UUID] = Field(None, title="User Id")
    is_read: Optional[bool] = Field(None, title="Is Read")


class NotificationReturning(BaseModel):
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    id: UUID = Field(..., title="Id")
    user_id: UUID = Field(..., title="User Id")
    type: str = Field(..., title="Type")
    message: str = Field(..., title="Message")
    is_read: bool = Field(..., title="Is Read")
    details: Dict[str, Any] = Field(..., title="Details")


class NotificationUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    is_read: Optional[bool] = Field(None, title="Is Read")
    message: Optional[str] = Field(None, title="Message")
    details: Optional[Dict[str, Union[str, int, float]]] = Field(None, title="Details")


class ObjectValidation(BaseModel):
    errors: Dict[str, List[str]] = Field(..., title="Errors")
    existing: Dict[str, List[str]] = Field(..., title="Existing")
    to_create: Dict[str, List[str]] = Field(..., title="To Create")


class OrgCreating(BaseModel):
    name: str = Field(..., title="Name")


class OrgDeleting(BaseModel):
    id: UUID = Field(..., title="Id")


class OrgReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    name: Optional[str] = Field(None, title="Name")


class OrgReturning(BaseModel):
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    id: UUID = Field(..., title="Id")
    name: str = Field(..., title="Name")


class OrgUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    name: Optional[str] = Field(None, title="Name")


class PackageCreating(BaseModel):
    name: str = Field(..., title="Name")
    filename: str = Field(..., title="Filename")
    broker_id: UUID = Field(..., title="Broker Id")
    version: str = Field(..., title="Version")
    environment: str = Field(..., title="Environment")
    meta: Optional[Dict[str, Any]] = Field({}, title="Meta")


class PackageDeleting(BaseModel):
    id: UUID = Field(..., title="Id")
    force: Optional[bool] = Field(False, title="Force")


class PackageListingLatest(BaseModel):
    org_id: Optional[UUID] = Field(None, title="Org Id")
    broker_id: Optional[UUID] = Field(None, title="Broker Id")


class PackageReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    name: Optional[str] = Field(None, title="Name")
    filename: Optional[str] = Field(None, title="Filename")
    org_id: Optional[UUID] = Field(None, title="Org Id")
    broker_id: Optional[UUID] = Field(None, title="Broker Id")
    version: Optional[str] = Field(None, title="Version")


class PackageReadingLatest(BaseModel):
    name: str = Field(..., title="Name")
    broker_id: UUID = Field(..., title="Broker Id")


class PackageReturning(BaseModel):
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    id: UUID = Field(..., title="Id")
    org_id: UUID = Field(..., title="Org Id")
    filename: str = Field(..., title="Filename")
    name: str = Field(..., title="Name")
    version: str = Field(..., title="Version")
    meta: Dict[str, Any] = Field(..., title="Meta")
    environment: str = Field(..., title="Environment")
    tasks_using: Optional[int] = Field(None, title="Tasks Using")
    actions_using: Optional[int] = Field(None, title="Actions Using")


class PackageSummary(BaseModel):
    id: UUID = Field(..., title="Id")
    name: str = Field(..., title="Name")
    version: str = Field(..., title="Version")


class PackageUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    meta: Optional[Dict[str, Any]] = Field(None, title="Meta")


class PageAssetReturning(BaseModel):
    items: List[AssetReturning] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PageBrokerPathReturning(BaseModel):
    items: List[BrokerPathReturning] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PageBrokerReturning(BaseModel):
    items: List[BrokerReturning] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PageDatasetReturning(BaseModel):
    items: List[DatasetReturning] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PageInvitationReturning(BaseModel):
    items: List[InvitationReturning] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PageNotificationReturning(BaseModel):
    items: List[NotificationReturning] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PageOrgReturning(BaseModel):
    items: List[OrgReturning] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PagePackageReturning(BaseModel):
    items: List[PackageReturning] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PersonCreating(BaseModel):
    name: str = Field(..., title="Name")
    source: str = Field(..., title="Source")
    email: EmailStr = Field(..., title="Email")
    picture: str = Field(..., title="Picture")
    locale: str = Field(..., title="Locale")
    avatar_url: str = Field(..., title="Avatar Url")
    preferences: Dict[str, Any] = Field(..., title="Preferences")


class PersonDeleting(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    email: Optional[str] = Field(None, title="Email")


class PersonReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    name: Optional[str] = Field(None, title="Name")
    source: Optional[str] = Field(None, title="Source")
    email: Optional[str] = Field(None, title="Email")
    verified: Optional[bool] = Field(None, title="Verified")


class PersonReturning(BaseModel):
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    id: UUID = Field(..., title="Id")
    name: str = Field(..., title="Name")
    source: str = Field(..., title="Source")
    email: str = Field(..., title="Email")
    picture: str = Field(..., title="Picture")
    avatar_url: str = Field(..., title="Avatar Url")
    locale: str = Field(..., title="Locale")
    verified: bool = Field(..., title="Verified")
    preferences: Dict[str, Any] = Field(..., title="Preferences")


class PersonUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    name: Optional[str] = Field(None, title="Name")
    source: Optional[str] = Field(None, title="Source")
    email: Optional[str] = Field(None, title="Email")
    picture: Optional[str] = Field(None, title="Picture")
    avatar_url: Optional[str] = Field(None, title="Avatar Url")
    locale: Optional[str] = Field(None, title="Locale")
    preferences: Optional[Dict[str, Any]] = Field(None, title="Preferences")
    verified: Optional[bool] = Field(None, title="Verified")
    last_push_time: Optional[datetime] = Field(None, title="Last Push Time")


class ProjectCreating(BaseModel):
    name: str = Field(..., min_length=3, title="Name")
    description: str = Field(..., title="Description")


class ProjectDeleting(BaseModel):
    id: UUID = Field(..., title="Id")


class ProjectReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    name: Optional[str] = Field(None, title="Name")
    org_id: Optional[UUID] = Field(None, title="Org Id")


class ProjectReturning(BaseModel):
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    id: UUID = Field(..., title="Id")
    org_id: UUID = Field(..., title="Org Id")
    name: str = Field(..., title="Name")
    description: str = Field(..., title="Description")


class ProjectStats(BaseModel):
    annotations: int = Field(..., title="Annotations")
    annotators: int = Field(..., title="Annotators")
    live_sessions: int = Field(..., title="Live Sessions")
    tasks_count: int = Field(..., title="Tasks Count")
    actions_count: int = Field(..., title="Actions Count")


class ProjectUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    name: Optional[str] = Field(None, title="Name")
    description: Optional[str] = Field(None, title="Description")


class RecipeDeleting(BaseModel):
    id: UUID = Field(..., title="Id")


class RecipeListingLatest(BaseModel):
    org_id: Optional[UUID] = Field(None, title="Org Id")
    broker_id: Optional[UUID] = Field(None, title="Broker Id")


class RecipePlanCreating(BaseModel):
    recipe_id: UUID = Field(..., title="Recipe Id")
    project_id: UUID = Field(..., title="Project Id")
    args: Dict[str, Any] = Field(..., title="Args")


class RecipePlanDeleting(BaseModel):
    id: UUID = Field(..., title="Id")


class RecipePlanReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    recipe_id: Optional[UUID] = Field(None, title="Recipe Id")
    project_id: Optional[UUID] = Field(None, title="Project Id")


class RecipePlanReturning(BaseModel):
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    id: UUID = Field(..., title="Id")
    recipe_id: UUID = Field(..., title="Recipe Id")
    project_id: UUID = Field(..., title="Project Id")
    environment: str = Field(..., title="Environment")
    recipe_name: str = Field(..., title="Recipe Name")
    args: Dict[str, Any] = Field(..., title="Args")
    command: List[str] = Field(..., title="Command")
    objects: Dict[str, Dict[str, Dict[str, Any]]] = Field(..., title="Objects")


class RecipePlanUpdating(BaseModel):
    id: UUID = Field(..., title="Id")


class RecipeReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    org_id: Optional[UUID] = Field(None, title="Org Id")
    name: Optional[str] = Field(None, title="Name")
    package_id: Optional[UUID] = Field(None, title="Package Id")


class RecipeReadingLatest(BaseModel):
    name: str = Field(..., title="Name")
    broker_id: UUID = Field(..., title="Broker Id")


class RecipeReturning(BaseModel):
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    id: UUID = Field(..., title="Id")
    org_id: UUID = Field(..., title="Org Id")
    package_id: UUID = Field(..., title="Package Id")
    name: str = Field(..., title="Name")
    title: str = Field(..., title="Title")
    description: str = Field(..., title="Description")
    is_action: bool = Field(..., title="Is Action")
    entry_point: str = Field(..., title="Entry Point")
    meta: Dict[str, Any] = Field(..., title="Meta")
    num_used_by: int = Field(..., title="Num Used By")


class RecipeUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    title: str = Field(..., title="Title")


class ReportNomadJobResponse(BaseModel):
    missing: List[UUID] = Field(..., title="Missing")


class SecretCreating(BaseModel):
    name: str = Field(..., title="Name")
    path: str = Field(..., title="Path")
    broker_id: UUID = Field(..., title="Broker Id")


class SecretDeleting(BaseModel):
    id: UUID = Field(..., title="Id")


class SecretDetail(BaseModel):
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    id: UUID = Field(..., title="Id")
    name: str = Field(..., title="Name")
    path: str = Field(..., title="Path")
    broker_id: UUID = Field(..., title="Broker Id")


class SecretReading(BaseModel):
    name: Optional[str] = Field(None, title="Name")
    path: Optional[str] = Field(None, title="Path")
    broker_id: Optional[UUID] = Field(None, title="Broker Id")


class SecretReturning(BaseModel):
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    id: UUID = Field(..., title="Id")
    name: str = Field(..., title="Name")
    path: str = Field(..., title="Path")
    broker_id: UUID = Field(..., title="Broker Id")


class SecretUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    name: str = Field(..., title="Name")
    path: str = Field(..., title="Path")


class SessionCreating(BaseModel):
    user_id: UUID = Field(..., title="User Id")
    task_id: UUID = Field(..., title="Task Id")


class SessionDeleting(BaseModel):
    id: UUID = Field(..., title="Id")


class SessionReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    user_id: Optional[UUID] = Field(None, title="User Id")
    task_id: Optional[UUID] = Field(None, title="Task Id")
    count: Optional[int] = Field(None, title="Count")
    time_spent: Optional[int] = Field(None, title="Time Spent")


class SessionReturning(BaseModel):
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    id: UUID = Field(..., title="Id")
    user_id: UUID = Field(..., title="User Id")
    user_name: str = Field(..., title="User Name")
    user_avatar_url: str = Field(..., title="User Avatar Url")
    task_id: UUID = Field(..., title="Task Id")
    count: int = Field(..., title="Count")
    time_spent: int = Field(..., title="Time Spent")


class SessionSaving(BaseModel):
    id: UUID = Field(..., title="Id")
    count: Optional[int] = Field(None, title="Count")
    time_spent: Optional[int] = Field(None, title="Time Spent")


class SessionUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    count: Optional[int] = Field(None, title="Count")


class TaskAnnotator(BaseModel):
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    id: UUID = Field(..., title="Id")
    org_id: UUID = Field(..., title="Org Id")
    org_name: str = Field(..., title="Org Name")
    person_id: UUID = Field(..., title="Person Id")
    name: str = Field(..., title="Name")
    email: str = Field(..., title="Email")
    avatar_url: str = Field(..., title="Avatar Url")
    locale: str = Field(..., title="Locale")
    groups: List[str] = Field(..., title="Groups")
    tags: List[str] = Field(..., title="Tags")
    is_org_admin: bool = Field(..., title="Is Org Admin")
    is_org_developer: bool = Field(..., title="Is Org Developer")
    live_sessions: int = Field(..., title="Live Sessions")
    annotations: int = Field(..., title="Annotations")
    hours: str = Field(..., title="Hours")


class TaskCreating(BaseModel):
    name: str = Field(..., title="Name")
    project_id: UUID = Field(..., title="Project Id")
    plan: RecipePlanCreating


class TaskDeleting(BaseModel):
    id: UUID = Field(..., title="Id")


class TaskEventDetails(BaseModel):
    fails_task: Optional[bool] = Field(False, title="Fails Task")
    restart_reason: Optional[str] = Field(None, title="Restart Reason")
    setup_error: Optional[str] = Field(None, title="Setup Error")
    driver_error: Optional[str] = Field(None, title="Driver Error")
    exit_code: Optional[int] = Field(None, title="Exit Code")
    signal: Optional[int] = Field(None, title="Signal")
    kill_timeout: Optional[str] = Field(None, title="Kill Timeout")
    kill_error: Optional[str] = Field(None, title="Kill Error")
    kill_reason: Optional[str] = Field(None, title="Kill Reason")
    start_delay: Optional[int] = Field(None, title="Start Delay")
    download_error: Optional[str] = Field(None, title="Download Error")
    validation_error: Optional[str] = Field(None, title="Validation Error")
    disk_limit: Optional[int] = Field(None, title="Disk Limit")
    failed_sibling: Optional[str] = Field(None, title="Failed Sibling")
    vault_error: Optional[str] = Field(None, title="Vault Error")
    task_signal_reason: Optional[str] = Field(None, title="Task Signal Reason")
    task_signal: Optional[str] = Field(None, title="Task Signal")


class TaskEventType(Enum):
    Alloc_Unhealthy = "Alloc Unhealthy"
    Setup_Failure = "Setup Failure"
    Driver_Failure = "Driver Failure"
    Received = "Received"
    Failed_Validation = "Failed Validation"
    Started = "Started"
    Terminated = "Terminated"
    Killing = "Killing"
    Killed = "Killed"
    Restarting = "Restarting"
    Not_Restarting = "Not Restarting"
    Restart_Signaled = "Restart Signaled"
    Signaling = "Signaling"
    Downloading_Artifacts = "Downloading Artifacts"
    Failed_Artifact_Download = "Failed Artifact Download"
    Building_Task_Directory = "Building Task Directory"
    Task_Setup = "Task Setup"
    Disk_Resources_Exceeded = "Disk Resources Exceeded"
    Sibling_Task_Failed = "Sibling Task Failed"
    Driver = "Driver"
    Leader_Task_Dead = "Leader Task Dead"
    Main_Tasks_Dead = "Main Tasks Dead"
    Task_hook_failed = "Task hook failed"
    Task_hook_message = "Task hook message"
    Failed_Restoring_Task = "Failed Restoring Task"
    Plugin_became_unhealthy = "Plugin became unhealthy"
    Plugin_became_healthy = "Plugin became healthy"
    Reconnected = "Reconnected"
    Waiting_for_shutdown_delay = "Waiting for shutdown delay"


class TaskPopulating(BaseModel):
    id: UUID = Field(..., title="Id")
    user_id: UUID = Field(..., title="User Id")


class TaskState(Enum):
    NOT_FOUND = "NOT_FOUND"
    CREATED = "CREATED"
    PREPARING = "PREPARING"
    WITH_PROBLEMS = "WITH_PROBLEMS"
    STARTING = "STARTING"
    STARTED = "STARTED"
    NOT_STARTED = "NOT_STARTED"
    STOPPING = "STOPPING"
    STOPPED = "STOPPED"
    DELETING = "DELETING"
    PENDING = "PENDING"
    FAILED = "FAILED"
    NOMAD_ERROR = "NOMAD_ERROR"
    NOMAD_JOB_ERROR = "NOMAD_JOB_ERROR"
    OBTAINING_MODELS_VENV = "OBTAINING_MODELS_VENV"
    INITIALIZING_VENV = "INITIALIZING_VENV"


class TaskStateReturning(BaseModel):
    updated: datetime = Field(..., title="Updated")
    id: UUID = Field(..., title="Id")
    state: TaskState
    is_running: bool = Field(..., title="Is Running")
    is_startable: bool = Field(..., title="Is Startable")
    is_stoppable: bool = Field(..., title="Is Stoppable")


class TaskStats(BaseModel):
    annotations: int = Field(..., title="Annotations")
    annotators: int = Field(..., title="Annotators")
    live_sessions: int = Field(..., title="Live Sessions")


class TaskSummary(BaseModel):
    id: UUID = Field(..., title="Id")
    project_id: UUID = Field(..., title="Project Id")
    name: str = Field(..., title="Name")
    job_type: JobType
    state: TaskState


class TaskUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    name: Optional[str] = Field(None, title="Name")


class TypeSchema(BaseModel):
    name: str = Field(..., title="Name")
    args: Optional[List[Union[str, TypeSchema]]] = Field(None, title="Args")
    object_type: Optional[str] = Field(None, title="Object Type")


class UserAuthenticated(BaseModel):
    access_token: str = Field(..., title="Access Token")
    expires_in: int = Field(..., title="Expires In")
    token_type: Optional[str] = Field("bearer", title="Token Type")


class UserCreating(BaseModel):
    org_id: UUID = Field(..., title="Org Id")
    person_id: UUID = Field(..., title="Person Id")
    is_org_admin: bool = Field(..., title="Is Org Admin")
    is_org_developer: bool = Field(..., title="Is Org Developer")
    tags: Optional[List[str]] = Field([], title="Tags")


class UserDeleting(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    org_id: Optional[UUID] = Field(None, title="Org Id")
    person_id: Optional[UUID] = Field(None, title="Person Id")


class UserReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    org_id: Optional[UUID] = Field(None, title="Org Id")
    person_id: Optional[UUID] = Field(None, title="Person Id")
    is_org_admin: Optional[bool] = Field(None, title="Is Org Admin")
    is_org_developer: Optional[bool] = Field(None, title="Is Org Developer")


class UserReturning(BaseModel):
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    id: UUID = Field(..., title="Id")
    org_id: UUID = Field(..., title="Org Id")
    org_name: str = Field(..., title="Org Name")
    person_id: UUID = Field(..., title="Person Id")
    name: str = Field(..., title="Name")
    email: str = Field(..., title="Email")
    avatar_url: str = Field(..., title="Avatar Url")
    locale: str = Field(..., title="Locale")
    groups: List[str] = Field(..., title="Groups")
    tags: List[str] = Field(..., title="Tags")
    is_org_admin: bool = Field(..., title="Is Org Admin")
    is_org_developer: bool = Field(..., title="Is Org Developer")


class UserUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    name: Optional[str] = Field(None, title="Name")
    email: Optional[str] = Field(None, title="Email")
    avatar_url: Optional[str] = Field(None, title="Avatar Url")
    locale: Optional[str] = Field(None, title="Locale")
    tags: Optional[List[str]] = Field([], title="Tags")


class ValidationError(BaseModel):
    loc: List[Union[str, int]] = Field(..., title="Location")
    msg: str = Field(..., title="Message")
    type: str = Field(..., title="Error Type")


class ValidationLevel(Enum):
    error = "error"
    warning = "warning"
    info = "info"
    success = "success"


class ValidationOp(Enum):
    ge = "ge"
    gt = "gt"
    le = "le"
    lt = "lt"
    eq = "eq"
    ne = "ne"
    re = "re"


class ValidationSchema(BaseModel):
    op: ValidationOp
    value: Union[int, float, str, bool] = Field(..., title="Value")
    message: str = Field(..., title="Message")
    level: ValidationLevel


class ActionCreating(BaseModel):
    project_id: UUID = Field(..., title="Project Id")
    name: str = Field(..., title="Name")
    evaluation: str = Field(..., title="Evaluation")
    plan: RecipePlanCreating


class ActionExecution(BaseModel):
    id: UUID = Field(..., title="Id")
    is_running: bool = Field(..., title="Is Running")
    started: datetime = Field(..., title="Started")
    runtime: Optional[float] = Field(None, title="Runtime")
    state: ActionState


class ActionReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    name: Optional[str] = Field(None, title="Name")
    state: Optional[ActionState] = None
    project_id: Optional[UUID] = Field(None, title="Project Id")
    broker_id: Optional[UUID] = Field(None, title="Broker Id")


class ActionReturning(BaseModel):
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    id: UUID = Field(..., title="Id")
    project_id: UUID = Field(..., title="Project Id")
    broker_id: UUID = Field(..., title="Broker Id")
    name: str = Field(..., title="Name")
    recipe_name: str = Field(..., title="Recipe Name")
    recipe_title: str = Field(..., title="Recipe Title")
    state: ActionState
    evaluation: str = Field(..., title="Evaluation")
    project_name: str = Field(..., title="Project Name")
    is_running: bool = Field(..., title="Is Running")
    is_startable: bool = Field(..., title="Is Startable")
    is_stoppable: bool = Field(..., title="Is Stoppable")
    error: Optional[str] = Field(None, title="Error")
    plan: RecipePlanReturning
    executions: List[ActionExecution] = Field(..., title="Executions")
    last_execution_id: Optional[UUID] = Field(None, title="Last Execution Id")
    nomad_index: int = Field(..., title="Nomad Index")
    stats: ActionStats
    cli_command: str = Field(..., title="Cli Command")
    url_logs: str = Field(..., title="Url Logs")
    url: str = Field(..., title="Url")
    related_tasks: List[UUID] = Field(..., title="Related Tasks")


class ActionSummary(BaseModel):
    id: UUID = Field(..., title="Id")
    project_id: UUID = Field(..., title="Project Id")
    name: str = Field(..., title="Name")
    job_type: JobType
    state: ActionState


class AssetDetail(BaseModel):
    created: Optional[datetime] = Field(None, title="Created")
    updated: Optional[datetime] = Field(None, title="Updated")
    id: UUID = Field(..., title="Id")
    broker_id: UUID = Field(..., title="Broker Id")
    name: str = Field(..., title="Name")
    kind: str = Field(..., title="Kind")
    path: str = Field(..., title="Path")
    meta: Dict[str, Any] = Field(..., title="Meta")
    tasks: List[TaskSummary] = Field(..., title="Tasks")
    actions: List[ActionSummary] = Field(..., title="Actions")


class ClusterJobReading(BaseModel):
    job_id: UUID = Field(..., title="Job Id")
    user_id: UUID = Field(..., title="User Id")
    job_type: JobType


class ClusterJobReturning(BaseModel):
    job_id: UUID = Field(..., title="Job Id")
    execution_id: Optional[UUID] = Field(None, title="Execution Id")
    attempt_id: UUID = Field(..., title="Attempt Id")
    job_type: JobType
    name: str = Field(..., title="Name")
    plan: RecipePlanReturning


class DatasetDetail(BaseModel):
    created: Optional[datetime] = Field(None, title="Created")
    updated: Optional[datetime] = Field(None, title="Updated")
    id: UUID = Field(..., title="Id")
    broker_id: UUID = Field(..., title="Broker Id")
    name: str = Field(..., title="Name")
    kind: str = Field(..., title="Kind")
    tasks: List[TaskSummary] = Field(..., title="Tasks")
    actions: List[ActionSummary] = Field(..., title="Actions")


class GroupReturning(BaseModel):
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    id: UUID = Field(..., title="Id")
    name: str = Field(..., title="Name")
    org_id: UUID = Field(..., title="Org Id")
    users: List[UserReturning] = Field(..., title="Users")


class HTTPValidationError(BaseModel):
    detail: Optional[List[ValidationError]] = Field(None, title="Detail")


class NomadEventBatchCreating(BaseModel):
    index: int = Field(..., title="Index")
    broker_id: UUID = Field(..., title="Broker Id")
    events: List[NomadEventCreating] = Field(..., title="Events")


class NomadTaskEvent(BaseModel):
    Type: TaskEventType
    Time: int = Field(..., title="Time")
    Message: str = Field(..., title="Message")
    DisplayMessage: str = Field(..., title="Displaymessage")
    Details: TaskEventDetails
    FailsTask: bool = Field(..., title="Failstask")
    RestartReason: str = Field(..., title="Restartreason")
    SetupError: str = Field(..., title="Setuperror")
    DriverError: str = Field(..., title="Drivererror")
    ExitCode: int = Field(..., title="Exitcode")
    Signal: int = Field(..., title="Signal")
    KillTimeout: int = Field(..., title="Killtimeout")
    KillError: str = Field(..., title="Killerror")
    KillReason: str = Field(..., title="Killreason")
    StartDelay: int = Field(..., title="Startdelay")
    DownloadError: str = Field(..., title="Downloaderror")
    ValidationError: str = Field(..., title="Validationerror")
    DiskLimit: int = Field(..., title="Disklimit")
    FailedSibling: str = Field(..., title="Failedsibling")
    VaultError: str = Field(..., title="Vaulterror")
    TaskSignalReason: str = Field(..., title="Tasksignalreason")
    TaskSignal: str = Field(..., title="Tasksignal")
    DriverMessage: str = Field(..., title="Drivermessage")
    GenericSource: str = Field(..., title="Genericsource")


class NomadTaskState(BaseModel):
    State: str = Field(..., title="State")
    Failed: bool = Field(..., title="Failed")
    Restarts: int = Field(..., title="Restarts")
    LastRestart: Optional[int] = Field(None, title="Lastrestart")
    StartedAt: Optional[datetime] = Field(None, title="Startedat")
    FinishedAt: Optional[datetime] = Field(None, title="Finishedat")
    Events: List[NomadTaskEvent] = Field(..., title="Events")


class PageActionReturning(BaseModel):
    items: List[ActionReturning] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PageGroupReturning(BaseModel):
    items: List[GroupReturning] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PagePersonReturning(BaseModel):
    items: List[PersonReturning] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PageProjectReturning(BaseModel):
    items: List[ProjectReturning] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PageRecipePlanReturning(BaseModel):
    items: List[RecipePlanReturning] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PageRecipeReturning(BaseModel):
    items: List[RecipeReturning] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PageSecretReturning(BaseModel):
    items: List[SecretReturning] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PageSessionReturning(BaseModel):
    items: List[SessionReturning] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PageTaskAnnotator(BaseModel):
    items: List[TaskAnnotator] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PageUserReturning(BaseModel):
    items: List[UserReturning] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class Props(BaseModel):
    name: Optional[str] = Field(None, title="Name")
    title: Optional[str] = Field(None, title="Title")
    description: Optional[str] = Field(None, title="Description")
    placeholder: Optional[str] = Field(None, title="Placeholder")
    value: Optional[Any] = Field(None, title="Value")
    required: Optional[bool] = Field(None, title="Required")
    positional: Optional[bool] = Field(None, title="Positional")
    exists: Optional[bool] = Field(None, title="Exists")
    validations: Optional[List[ValidationSchema]] = Field(None, title="Validations")
    pattern: Optional[str] = Field(None, title="Pattern")
    widget: Optional[str] = Field(None, title="Widget")
    choice: Optional[Dict[str, Props]] = Field(None, title="Choice")
    min: Optional[float] = Field(None, title="Min")
    max: Optional[float] = Field(None, title="Max")
    step: Optional[float] = Field(None, title="Step")
    optional_title: Optional[str] = Field(None, title="Optional Title")


class TaskReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    name: Optional[str] = Field(None, title="Name")
    state: Optional[TaskState] = None
    project_id: Optional[UUID] = Field(None, title="Project Id")


class TaskReturning(BaseModel):
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    id: UUID = Field(..., title="Id")
    project_id: UUID = Field(..., title="Project Id")
    broker_id: UUID = Field(..., title="Broker Id")
    name: str = Field(..., title="Name")
    recipe_name: str = Field(..., title="Recipe Name")
    recipe_title: str = Field(..., title="Recipe Title")
    state: TaskState
    project_name: str = Field(..., title="Project Name")
    is_running: bool = Field(..., title="Is Running")
    is_startable: bool = Field(..., title="Is Startable")
    is_stoppable: bool = Field(..., title="Is Stoppable")
    error: Optional[str] = Field(None, title="Error")
    plan: RecipePlanReturning
    nomad_index: int = Field(..., title="Nomad Index")
    last_execution_id: Optional[UUID] = Field(None, title="Last Execution Id")
    related_actions: List[UUID] = Field(..., title="Related Actions")
    cli_command: str = Field(..., title="Cli Command")
    url_logs: str = Field(..., title="Url Logs")
    url: str = Field(..., title="Url")


class FieldSchema(BaseModel):
    name: str = Field(..., title="Name")
    type: TypeSchema
    props: Props
    fields: Optional[List[FieldSchema]] = Field([], title="Fields")
    choice: Optional[List[FieldSchema]] = Field([], title="Choice")


class NomadAllocationStub(BaseModel):
    ID: UUID = Field(..., title="Id")
    EvalID: UUID = Field(..., title="Evalid")
    Name: str = Field(..., title="Name")
    Namespace: str = Field(..., title="Namespace")
    NodeID: UUID = Field(..., title="Nodeid")
    NodeName: str = Field(..., title="Nodename")
    JobID: UUID = Field(..., title="Jobid")
    JobType: Optional[str] = Field(None, title="Jobtype")
    JobVersion: Optional[int] = Field(None, title="Jobversion")
    TaskGroup: str = Field(..., title="Taskgroup")
    DesiredStatus: AllocDesiredStatus
    DesiredDescription: Optional[str] = Field(None, title="Desireddescription")
    ClientStatus: AllocClientStatus
    ClientDescription: Optional[str] = Field(None, title="Clientdescription")
    TaskStates: Optional[Dict[str, NomadTaskState]] = Field({}, title="Taskstates")
    FollowupEvalID: Optional[UUID] = Field(None, title="Followupevalid")
    PreemptedAllocations: Optional[List[str]] = Field(
        None, title="Preemptedallocations"
    )
    PreemptedByAllocation: Optional[UUID] = Field(None, title="Preemptedbyallocation")
    CreateIndex: int = Field(..., title="Createindex")
    ModifyIndex: int = Field(..., title="Modifyindex")
    CreateTime: int = Field(..., title="Createtime")
    ModifyTime: int = Field(..., title="Modifytime")


class PageTaskReturning(BaseModel):
    items: List[TaskReturning] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class RecipeSchema(BaseModel):
    name: str = Field(..., title="Name")
    fields: List[FieldSchema] = Field(..., title="Fields")
    title: Optional[str] = Field(None, title="Title")
    description: Optional[str] = Field(None, title="Description")
    view_id: Optional[str] = Field(None, title="View Id")
    cli_names: Optional[Dict[str, str]] = Field(None, title="Cli Names")


class NomadJobs(BaseModel):
    jobs: List[NomadJobStub] = Field(..., title="Jobs")
    allocations: List[NomadAllocationStub] = Field(..., title="Allocations")


class RecipeCreating(BaseModel):
    name: str = Field(..., title="Name")
    title: str = Field(..., title="Title")
    description: str = Field(..., title="Description")
    is_action: bool = Field(..., title="Is Action")
    entry_point: str = Field(..., title="Entry Point")
    form_schema: RecipeSchema
    cli_schema: Dict[str, Any] = Field(..., title="Cli Schema")
    meta: Dict[str, Any] = Field(..., title="Meta")
    package_id: UUID = Field(..., title="Package Id")


class RecipeDetail(BaseModel):
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    id: UUID = Field(..., title="Id")
    org_id: UUID = Field(..., title="Org Id")
    package_id: UUID = Field(..., title="Package Id")
    name: str = Field(..., title="Name")
    title: str = Field(..., title="Title")
    description: str = Field(..., title="Description")
    is_action: bool = Field(..., title="Is Action")
    entry_point: str = Field(..., title="Entry Point")
    form_schema: RecipeSchema
    cli_schema: Dict[str, Any] = Field(..., title="Cli Schema")
    meta: Dict[str, Any] = Field(..., title="Meta")
    tasks: List[TaskSummary] = Field(..., title="Tasks")
    actions: List[ActionSummary] = Field(..., title="Actions")
    package: Optional[PackageSummary] = None
    num_used_by: int = Field(..., title="Num Used By")


CackQuerying.update_forward_refs()
TypeSchema.update_forward_refs()
Props.update_forward_refs()
FieldSchema.update_forward_refs()



from ..models import (
    SessionCreating,
    SessionDeleting,
    SessionReading,
    SessionReturning,
    SessionUpdating,
)
from .base import ModelClient


class Session(
    ModelClient[
        SessionCreating,
        SessionReading,
        SessionUpdating,
        SessionDeleting,
        SessionReturning,
        SessionReturning,
    ]
):
    Creating = SessionCreating
    Reading = SessionReading
    Updating = SessionUpdating
    Deleting = SessionDeleting
    Returning = SessionReturning



import re
from datetime import datetime
from enum import Enum
from pathlib import Path
from types import ModuleType
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Generator,
    Generic,
    Iterator,
    List,
    Literal,
    NamedTuple,
    NoReturn,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)
from uuid import UUID

from pydantic import BaseModel, ConstrainedStr, Field, conint, constr
from pydantic.generics import GenericModel


class NotificationTypes(str, Enum):
    """Types of notifications.

    Should match the values in app/util/constants.ts"""

    invite_accept = "INVITE_ACCEPT"
    system = "SYSTEM"
    goal = "GOAL"


T = TypeVar("T")


class Page(GenericModel, Generic[T]):
    items: Sequence[T]
    total: int
    page: conint(ge=1)  # type: ignore
    size: conint(ge=1)  # type: ignore


class Fields:
    str255: str = Field(..., max_length=255)
    maybe_str255: Optional[str] = Field(None, lt=255)


SEMVER_PATTERN = re.compile(
    r"^((([0-9]+)\.([0-9]+)\.([0-9]+)(?:-([0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?)(?:\+([0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?)$"  # noqa
)


class SemverStr(ConstrainedStr):
    regex = SEMVER_PATTERN


FuncT = TypeVar("FuncT", bound=Callable)
Wrapper = Callable[[FuncT], FuncT]
Decorator = Callable[[Any, Any], Wrapper]


__all__ = [
    "Any",
    "Callable",
    "Union",
    "List",
    "Dict",
    "NoReturn",
    "Optional",
    "Wrapper",
    "Decorator",
    "Sequence",
    "cast",
    "Iterator",
    "Set",
    "Tuple",
    "Type",
    "TypeVar",
    "Generic",
    "ClassVar",
    "Path",
    "Generator",
    "datetime",
    "constr",
    "Fields",
    "BaseModel",
    "ModuleType",
    "UUID",
    "Literal",
    "NamedTuple",
]



import json
import os
import re
import shutil
import tarfile
import time
from dataclasses import dataclass
from pathlib import Path
from zipfile import ZipFile

import packaging.tags
import packaging.utils
from radicli import Arg
from wasabi import msg

from .. import ty
from ..auth import get_current_auth
from ..cli import cli
from ..errors import BrokerError, CLIError
from ..messages import Messages
from ..prodigy_teams_broker_sdk import Client as BrokerClient
from ..prodigy_teams_broker_sdk import models as broker_models
from ..prodigy_teams_pam_sdk import Client
from ..prodigy_teams_pam_sdk.models import (
    PackageCreating,
    PackageReading,
    RecipeCreating,
    RecipeDetail,
    RecipeListingLatest,
    RecipeReturning,
)
from ..query import resolve_recipe
from ..ui import dicts_to_table, print_info_table, print_table_with_select
from ..util import resolve_remote_path
from ._recipes_meta import build_wheel, compute_recipes_meta

COOKIECUTTER_PATH = Path(__file__).parent.parent / "recipes_cookiecutter"


@cli.subcommand("recipes", "list")
def list() -> ty.Sequence[RecipeReturning]:
    """List all recipes"""
    auth = get_current_auth()
    # When there are multiple versions of a package publishing some recipe, only
    # list the latest version.
    res = auth.client.recipe.all_latest(
        body=RecipeListingLatest(broker_id=auth.broker_id, org_id=auth.org_id)
    )
    print_table_with_select(res.items, ["id", "name"])
    return res.items


@cli.subcommand(
    "recipes",
    "info",
    name_or_id=Arg(help=Messages.name_or_id.format(noun="recipe")),
    cluster_id=Arg(help=Messages.cluster_id.format(noun="recipe")),
    as_json=Arg("--json", help=Messages.as_json),
)
def info(
    name_or_id: ty.StrOrUUID,
    cluster_id: ty.Optional[ty.UUID] = None,
    as_json: bool = False,
) -> RecipeDetail:
    """Show info about a recipe"""
    auth = get_current_auth()
    if cluster_id is None:
        cluster_id = auth.broker_id
    res = resolve_recipe(name_or_id, broker_id=cluster_id)
    print_info_table(res, as_json=as_json)
    return res


@cli.subcommand(
    "recipes",
    "init",
    # fmt: off
    output_dir=Arg(help=Messages.output_dir.format(noun="recipe package")),
    name=Arg("--name", help=Messages.name.format(noun="package (e.g. custom_recipes)")),
    version=Arg("--version", help=Messages.version.format(noun="package")),
    description=Arg("--description", help=Messages.description.format(noun="package")),
    author=Arg("--author", help=Messages.name.format(noun="package author")),
    email=Arg("--email", help=Messages.email.format(noun="package author")),
    url=Arg("--url", help=Messages.url.format(noun="package")),
    license=Arg("--license", help=Messages.license.format(noun="package")),
    # fmt: on
)
def init(
    output_dir: Path,
    name: ty.Optional[str] = None,
    version: str = "0.1.0",
    description: str = "",
    author: str = "",
    email: str = "",
    url: str = "",
    license: str = "",
) -> None:
    """Generate a new recipes Python package"""
    # output_dir is passed as the path to the package itself, but our template
    # structure includes the package directory under the `package_dir` key.
    package_dir = output_dir.resolve()
    parent_dir = package_dir.parent.resolve()
    package_dir = package_dir.name
    # Infer a default package name from the directory path
    name = name or package_dir
    # Create a wheel-friendly package name slug
    package_name = re.sub(r"[^\w\d.]+", "_", name.lower(), re.UNICODE)
    template_dir = COOKIECUTTER_PATH
    variables = {
        "name": name,  # human friendly recipe package name
        "version": version,
        "short_description": description,
        "author": author,
        "email": email,
        "url": url,
        "license": license,
        "parent_dir": str(parent_dir),  # the directory path containing the package
        "package_dir": package_dir,  # top-level package directory within `parent_dir`
        "package_name": package_name,  # the normalized name of the package
    }
    if not _ensure_preconditions(template_dir, variables):
        msg.info(Messages.E154.format(name=name))
        return None
    _fill_template(template_dir, variables)
    msg.good(Messages.T002.format(noun="package", name=name), str(output_dir))


@cli.subcommand(
    "recipes",
    "verify",
    package=Arg(help=Messages.path.format(noun="package")),
)
def verify(package: ty.ExistingFilePath) -> None:
    """Verify a built recipe package before upload"""
    if not (package.name.endswith("tar.gz") or package.name.endswith(".whl")):
        raise CLIError(Messages.E144, package.name)
    file_name = ""
    meta_json = None
    valid_meta = False
    if package.name.endswith(".whl"):
        with ZipFile(package) as zip_file:
            for file_name in zip_file.namelist():
                if file_name.endswith("/meta.json"):
                    meta_bytes = zip_file.read(file_name)
                    meta_json = json.loads(meta_bytes.decode("utf-8"))
                    if not isinstance(meta_json, dict):
                        continue
                    if "prodigy_teams" in meta_json:
                        valid_meta = True
                        break
    elif package.name.endswith(".tar.gz"):
        with tarfile.open(package) as tar:
            for file_name in tar.getnames():
                if file_name.endswith("/meta.json"):
                    io_bytes = tar.extractfile(file_name)
                    assert io_bytes
                    meta_json = json.load(io_bytes)
                    if not isinstance(meta_json, dict):
                        continue
                    if "prodigy_teams" in meta_json:
                        valid_meta = True
                        break
    if not (valid_meta and meta_json):
        raise CLIError(Messages.E145)
    msg.good(Messages.T022, file_name)
    recipes_data = meta_json.get("prodigy_teams", {}).get("recipes")
    if not isinstance(recipes_data, dict):
        raise CLIError(Messages.E146, recipes_data)
    if not recipes_data:
        raise CLIError(Messages.E147, Messages.E148)
    recipes_info = []
    invalid_recipes = []
    for key, data in recipes_data.items():
        entry_point = data.get("entry_point")
        recipe_args = data.get("args")
        data = {
            "name": key,
            "entry_point": entry_point,
            "contains args": bool(recipe_args),
        }
        recipes_info.append(data)
        if not entry_point or not recipe_args:
            invalid_recipes.append(key)
    headers, rows = dicts_to_table(recipes_info)
    msg.good(Messages.T023)
    msg.table(rows, header=headers, divider=True, max_col=3000)
    if invalid_recipes:
        raise CLIError(Messages.E149, ", ".join(invalid_recipes))
    msg.good(Messages.T024)


@cli.subcommand(
    "recipes",
    "publish",
    # fmt: off
    src=Arg(help=Messages.local_path.format(noun="package")),
    wheelhouse=Arg(help=Messages.wheelhouse),
    cluster_wheels_path=Arg("--cluster-wheels-path", help=Messages.path.format(noun="cluster wheels")),  # TODO
    cluster_envs_path=Arg("--cluster-envs-path", help=Messages.path.format(noun="cluster envs")),  # TODO
    use_active_venv=Arg("--use-active-venv", help=Messages.use_active_venv),
    exists_ok=Arg("--exists-ok", help=Messages.exists_ok)
    # fmt: on
)
def publish(
    src: ty.ExistingDirPath,
    wheelhouse: Path,
    cluster_wheels_path: str = "{wheels}",
    cluster_envs_path: str = "{envs}",
    use_active_venv: bool = False,
    exists_ok: bool = False,
) -> ty.Optional[ty.UUID]:
    """
    Build, upload and advertise a recipes package from the local filesystem. The
    recipes package and any required dependencies are uploaded to the cluster,
    and then advertised to the PAM service.
    """
    auth = get_current_auth()
    if auth.broker_host is None:
        raise CLIError(Messages.E035)
    if auth.broker_id is None:
        raise CLIError(Messages.E036)
    # Before starting to build the wheel, do some work that could error. Building
    # the wheel and getting the metadata is slow, so we want to avoid it if possible.
    cluster_wheels_path = resolve_remote_path(
        auth.client, cluster_wheels_path, auth.broker_host
    )
    cluster_envs_path = resolve_remote_path(
        auth.client, cluster_envs_path, auth.broker_host
    )
    # Sync wheels to broker. We first calculate what files the broker needs to
    # fulfill the specified requirements.
    body = broker_models.Listing(
        path=cluster_wheels_path, recurse=True, include_stats=True
    )
    try:
        cluster_res = auth.broker_client.files.list_dir(body)
    except BrokerError as e:
        raise CLIError(Messages.E152, e)
    else:
        assert cluster_res.stats is not None
        cluster_wheels = {}
        for path, stats in zip(cluster_res.paths, cluster_res.stats):
            cluster_wheels[Path(path).name] = stats.size
    if src.is_dir():
        msg.info(Messages.T025.format(src=src))
        wheel = build_wheel(src, wheelhouse, use_active_venv)
    else:
        wheel = wheelhouse / src.name
        if src != wheel:
            shutil.copyfile(src, wheel)
    package_exists = _check_for_package(auth.client, auth.broker_id, wheel)
    if package_exists:
        if exists_ok:
            msg.info(Messages.T026.format(name=wheel.name))
            return None
        raise CLIError(Messages.E153.format(name=wheel.name))
    # Builds the package
    msg.info(Messages.T027.format(name=wheel.name))
    meta = compute_recipes_meta(wheel, wheelhouse, use_active_venv)
    msg.good(Messages.T028)
    msg.info(Messages.T029)
    to_upload = _plan_wheel_upload(
        wheel=wheel,
        requirements=meta["requirements"],
        wheelhouse=wheelhouse,
        cluster_wheels=cluster_wheels,
    )
    msg.info(Messages.T030.format(count=len(to_upload)))
    # Upload the wheels
    _upload_wheels(auth.broker_client, to_upload, cluster_wheels_path)
    msg.good(Messages.T028)
    # Create the environment. First we try non-blocking, but if that fails, we
    # try again with blocking to (hopefully!) surface the error.
    msg.info(Messages.T031)
    maybe_env_path = _create_environment(
        auth.broker_client,
        meta["requirements"],
        wheels_path=cluster_wheels_path,
        envs_path=cluster_envs_path,
        blocking=False,
    )
    if maybe_env_path is None:
        env_path = _create_environment(
            auth.broker_client,
            meta["requirements"],
            wheels_path=cluster_wheels_path,
            envs_path=cluster_envs_path,
            blocking=True,
        )
        assert env_path is not None
    else:
        env_path = maybe_env_path
    msg.good(Messages.T028)
    msg.info(Messages.T032)
    package_id = _publish_package_to_pam(
        auth.client,
        auth.broker_id,
        name=meta["name"],
        filename=wheel.name,
        version=meta["version"],
        meta=meta,
        environment=env_path,
    )
    msg.good(Messages.T028)
    return package_id


def _check_for_package(client: Client, broker_id: ty.UUID, wheel_path) -> bool:
    packages = client.package.all(
        PackageReading(
            broker_id=broker_id,
            filename=wheel_path.name,
            name=None,
            version=None,
            id=None,
            org_id=None,
        )
    )
    return bool(packages.items)


def _plan_wheel_upload(
    *,
    wheel: Path,
    requirements: ty.Set[str],
    wheelhouse: Path,
    cluster_wheels: ty.Dict[str, int],
) -> ty.List[Path]:
    needed = [wheel]
    for path in wheelhouse.iterdir():
        if path.name[0] == "." or path.suffix != ".whl":
            continue
        if (
            path.name in cluster_wheels
            and cluster_wheels[path.name] == path.stat().st_size
        ):
            continue
        if path.name == wheel.name:
            continue
        wheel_info = _WheelInfo.from_filename(path.name)
        if f"{wheel_info.name}=={wheel_info.version}" not in requirements:
            continue
        # TODO compatibility check
        # TODO check whether on public pypi?
        needed.append(path.absolute())
    return needed


def _upload_wheels(client: BrokerClient, srcs: ty.List[Path], dest: str) -> None:
    for src in srcs:
        with src.open("rb") as file_:
            client.files.upload(
                file_,
                dest=os.path.join(dest, src.name),
                make_dirs=True,
                overwrite=True,
            )
            msg.good(Messages.T033.format(src=src.name))


def _create_environment(
    client: BrokerClient,
    requirements: ty.List[str],
    wheels_path: str,
    envs_path: str,
    blocking: bool,
) -> ty.Optional[str]:
    body = broker_models.EnvCreating(
        requirements=requirements,
        wheels_path=wheels_path,
        envs_path=envs_path,
        blocking=blocking,
    )
    env = client.envs2.create(body)
    if blocking:
        # It's supposed to only return when the environment is ready if we told it to block
        assert env.ready
    else:
        for _ in range(50):
            if env.ready:
                return env.path
            else:
                time.sleep(10)
                env = client.envs2.await_env(broker_models.EnvAwaiting(path=env.path))
        return None


def _publish_package_to_pam(
    client: Client,
    broker_id: ty.UUID,
    *,
    name: str,
    filename: str,
    version: str,
    environment: str,
    meta: ty.Dict[str, ty.Any],
) -> ty.UUID:
    package = client.package.create(
        PackageCreating(
            name=name,
            filename=filename,
            version=version,
            broker_id=broker_id,
            environment=environment,
            meta=meta,
        )
    )
    msg.good(
        Messages.T002.format(noun="package", name=f"{package.name} ({package.id})")
    )
    for recipe_data in meta["recipes"]:
        # The recipe create-meta command is supposed to output
        # entries that match the RecipeCreating body, except
        # for missing a package_id
        body = RecipeCreating(**recipe_data, package_id=package.id)
        r = client.recipe.create(body)
        recipe_type = "action" if body.is_action is True else "task"
        msg.good(Messages.T002.format(noun=f"{recipe_type} recipe", name=r.name))
    return package.id


def _ensure_preconditions(template_dir: Path, variables: ty.Dict[str, ty.Any]) -> bool:
    output_dir = Path(variables["parent_dir"])
    if output_dir.exists() and output_dir.is_file():
        msg.fail(Messages.E002.format(noun="directory", name=str(output_dir)))
        return False
    conflicts = _find_overwrite_conflicts(template_dir, variables)
    if conflicts:
        for conflict_file in conflicts:
            msg.fail(Messages.E002.format(noun="file", name=str(conflict_file)))
        return False
    return True


def _find_overwrite_conflicts(
    template_dir: Path, variables: ty.Dict[str, ty.Any]
) -> ty.List[ty.Union[Path, "TemplateItem"]]:
    output_dir = Path(variables["parent_dir"])
    conflicts = []
    if output_dir.exists() and output_dir.is_dir():
        for tmpl in _walk_template(template_dir):
            out_file = tmpl.output_path(variables)
            if out_file.exists():
                conflicts.append(out_file)
    elif output_dir.exists() and output_dir.is_file():
        conflicts.append(output_dir)
    return conflicts


@dataclass
class TemplateItem:
    template_dir: Path
    template_path: Path

    def output_path(self, variables: ty.Dict[str, ty.Any]) -> Path:
        return Path(variables["parent_dir"]) / _replace_path_vars(
            self.template_path.relative_to(self.template_dir), variables
        )

    def expand(self, variables: ty.Dict[str, ty.Any]) -> str:
        with self.template_path.open("r", encoding="utf8") as file_:
            return _replace_content_vars(file_.read(), variables)


def _fill_template(template_dir: Path, variables: ty.Dict[str, ty.Any]) -> None:
    # TODO: check cookiecutter.json for defaults?
    for tmpl in _walk_template(template_dir):
        contents = tmpl.expand(variables)
        output_path = tmpl.output_path(variables)
        if output_path.name.endswith(".tmpl"):
            # We allow files to be named *.thing.tmpl to denote that it's a template
            # of a .thing file, but not currently a valid .thing file. We still might
            # replace variables in files named other things, if the variable substitution
            # doesn't change the syntactic validity of the file. So if we have a .tmpl
            # suffix, we just strip that.
            output_path = output_path.parent / output_path.name[: -len(".tmpl")]
        output_path.parent.mkdir(exist_ok=True, parents=True)
        with output_path.open("w", encoding="utf8") as file_:
            file_.write(contents)


def _replace_content_vars(
    contents: str,
    variables: ty.Dict[str, ty.Any],
    *,
    var_prefix="{{cookiecutter.",
    var_suffix="}}",
) -> str:
    for key, value in variables.items():
        key = f"{var_prefix}{key}{var_suffix}"
        contents = contents.replace(key, value)
    return contents


def _replace_path_vars(
    path: Path,
    variables: ty.Dict[str, ty.Any],
    *,
    var_prefix="{{cookiecutter.",
    var_suffix="}}",
) -> Path:
    """Given a path that might have variables, fill in the variables
    from a given dict.
    """
    parts = []
    for part in path.parts:
        if part.startswith(var_prefix) and part.endswith(var_suffix):
            variable = part[len(var_prefix) : -len(var_suffix)]
            part = variables[variable]
        parts.append(part)
    if not parts:
        return path
    else:
        return Path(parts[0]).joinpath(*parts[1:])


def _walk_template(template_dir: Path) -> ty.Iterable[TemplateItem]:
    def _walk(path: Path) -> ty.Iterable[Path]:
        for p in Path(path).iterdir():
            if p.is_dir():
                yield from _walk(p)
                continue
            if p.name == "cookiecutter.json":
                continue
            yield p

    for p in _walk(template_dir):
        yield TemplateItem(template_dir, p)


@dataclass
class _WheelInfo:
    filename: str
    name: str
    version: str
    build: ty.Optional[ty.Tuple[int, str]]
    tags: ty.Set[packaging.tags.Tag]

    @classmethod
    def from_filename(cls, filename: str) -> "_WheelInfo":
        name, version, build, tags = packaging.utils.parse_wheel_filename(filename)
        return cls(
            filename=filename,
            name=str(name),
            version=str(version),
            build=build if build else None,
            tags=set(tags),
        )



from .. import ty
from ..models import EnvAwaiting, EnvCreating, EnvReturning
from .base import BaseClient


class Envs2(BaseClient):
    def create(self, body: EnvCreating) -> EnvReturning:
        res = self.request(
            "POST", endpoint="create", data=body, return_model=EnvReturning
        )
        return ty.cast(EnvReturning, res)

    async def create_async(self, body: EnvCreating) -> EnvReturning:
        res = self.request_async(
            "POST", endpoint="create", data=body, return_model=EnvReturning
        )
        return ty.cast(EnvReturning, res)

    def await_env(self, body: EnvAwaiting) -> EnvReturning:
        res = self.request(
            "POST", endpoint="await", data=body, return_model=EnvReturning
        )
        return ty.cast(EnvReturning, res)

    async def await_env_async(self, body: EnvAwaiting) -> EnvReturning:
        res = self.request_async(
            "POST", endpoint="await", data=body, return_model=EnvReturning
        )
        return ty.cast(EnvReturning, res)



from radicli import Arg
from wasabi import msg

from .. import ty
from ..auth import get_current_auth
from ..cli import cli
from ..errors import CLIError, ProdigyTeamsErrors
from ..messages import Messages
from ..prodigy_teams_pam_sdk.models import BrokerPathReturning
from ..query import resolve_path, resolve_path_id
from ..ui import print_info_table, print_table_with_select
from ..util import resolve_remote_path


@cli.subcommand("paths", "list")
def list() -> ty.Sequence[BrokerPathReturning]:
    """List all cluster path aliases"""
    client = get_current_auth().client
    res = client.broker_path.all()
    print_table_with_select(res.items, select=["created", "id", "name", "path"])
    return res.items


@cli.subcommand(
    "paths",
    "info",
    name_or_id=Arg(help=Messages.name_or_id.format(noun="path")),
    cluster_id=Arg(help=Messages.cluster_id.format(noun="path")),
    as_json=Arg("--json", help=Messages.as_json),
)
def info(
    name_or_id: ty.StrOrUUID,
    cluster_id: ty.Optional[ty.UUID] = None,
    as_json: bool = False,
) -> BrokerPathReturning:
    """Get detailed info for a path alias"""
    res = resolve_path(name_or_id, broker_id=cluster_id)
    print_info_table(res, as_json=as_json)
    return res


@cli.subcommand(
    "paths",
    "create",
    name=Arg(help=Messages.name.format(noun="path")),
    path=Arg(help=Messages.path.format(noun="cluster directory")),
    exists_ok=Arg("--exists-ok", help=Messages.exists_ok),
)
def create(name: str, path: str, exists_ok: bool = False) -> ty.Optional[ty.UUID]:
    """Create a new path alias"""
    auth = get_current_auth()
    client = auth.client
    broker_id = auth.broker_id
    path = resolve_remote_path(client, path, default_broker=auth.broker_host)
    try:
        res = client.broker_path.create(name=name, path=path, broker_id=broker_id)
    except ProdigyTeamsErrors.BrokerPathExists:
        if exists_ok:
            msg.info(Messages.T001.format(noun="path", name=name))
            return None
        raise CLIError(Messages.E002.format(noun="path", name=name))
    except ProdigyTeamsErrors.BrokerPathInvalid:
        raise CLIError(Messages.E004.format(noun="path", name=name))
    except ProdigyTeamsErrors.BrokerPathForbiddenCreate:
        raise CLIError(Messages.E003.format(noun="path", name=name))
    msg.divider("Path Alias")
    msg.table(res.dict())
    return res.id


@cli.subcommand(
    "paths",
    "delete",
    name_or_id=Arg(help=Messages.name_or_id.format(noun="path")),
    cluster_id=Arg(help=Messages.cluster_id.format(noun="path")),
)
def delete(
    name_or_id: ty.StrOrUUID, cluster_id: ty.Optional[ty.UUID] = None
) -> ty.UUID:
    """Delete a path alias"""
    path_id = resolve_path_id(name_or_id, broker_id=cluster_id)
    auth = get_current_auth()
    try:
        auth.client.broker_path.delete(id=path_id)
    except (
        ProdigyTeamsErrors.ProjectForbiddenDelete,
        ProdigyTeamsErrors.ProjectNotFound,
    ):
        raise CLIError(Messages.E006.format(noun="path", name=name_or_id))
    else:
        msg.good(Messages.T003.format(noun="path", name=name_or_id))
    return path_id



import sys

from radicli import Arg
from wasabi import msg

from .. import ty
from ..auth import get_current_auth
from ..cli import cli
from ..errors import BrokerError, CLIError, HTTPXErrors, ProdigyTeamsError
from ..messages import Messages
from ..prodigy_teams_broker_sdk.models import JobLogRequest
from ..prodigy_teams_pam_sdk.models import ActionReturning
from ..query import delete_job, resolve_action, resolve_recipe, start_job, stop_job
from ..ui import (
    print_args_table,
    print_info_table,
    print_logs,
    print_recipes_help,
    print_table_with_select,
)
from ._recipe_subcommand import create_from_recipe, request_recipes


@cli.subcommand_with_extra(
    "actions",
    "create",
    exists_ok=Arg("--exists-ok", help=Messages.exists_ok),
    no_start=Arg("--no-start", help=Messages.no_start),
    _show_help=Arg("--help", "-h", help=Messages.help),
)
def create(
    exists_ok: bool = False,
    no_start: bool = False,
    _show_help: bool = False,
    _extra: ty.List[str] = [],
) -> ty.UUID:
    """
    Create a new action. The available action recipes are fetched from your
    cluster and are added as dynamic subcommands. You can see more details
    and available arguments by calling the subcommand with --help, e.g. create
    [name] --help
    """
    auth = get_current_auth()
    try:
        action_schemas = request_recipes(auth=auth, is_action=True)
    except ProdigyTeamsError as e:
        raise CLIError(Messages.E009, e)
    except HTTPXErrors as e:
        raise CLIError(Messages.E009, e)
    if not _extra:
        print_recipes_help(action_schemas, "Create a new action", "actions create")
        sys.exit(0)
    args = [*_extra]
    name = args.pop(0)
    if name not in action_schemas:
        opts = f"Available: {', '.join(action_schemas.keys())}"
        raise CLIError(Messages.E010.format(noun="action", name=name), opts)
    schema = resolve_recipe(name, broker_id=auth.broker_id)
    action_id, plan = create_from_recipe(
        schema, args, command="actions", show_help=_show_help, exists_ok=exists_ok
    )
    msg.good(Messages.T002.format(noun="action", name=action_id))
    print_args_table(plan.args, schema.form_schema.cli_names)
    if not no_start:
        start(action_id)
    return action_id


@cli.subcommand(
    "actions",
    "list",
    # fmt: off
    select=Arg("--select", help=Messages.select.format(opts=list(ActionReturning.__fields__))),
    # fmt: on
)
def list(
    select: ty.List[str] = ["id", "name", "state"]
) -> ty.Sequence[ActionReturning]:
    """
    List the actions on the cluster. By default, this includes their ID, name
    and current state, e.g. created or completed
    """
    client = get_current_auth().client
    res = client.action.all(page=1, size=100)
    print_table_with_select(res.items, select=select)
    return res.items


@cli.subcommand(
    "actions",
    "info",
    name_or_id=Arg(help=Messages.name_or_id.format(noun="action")),
    project_id=Arg(help=Messages.project_id.format(noun="action")),
    cluster_id=Arg(help=Messages.cluster_id.format(noun="action")),
    as_json=Arg("--json", help=Messages.as_json),
)
def info(
    name_or_id: ty.StrOrUUID,
    cluster_id: ty.Optional[ty.UUID] = None,
    project_id: ty.Optional[ty.UUID] = None,
    as_json: bool = False,
) -> ActionReturning:
    """Print information about an action on the cluster"""
    res = resolve_action(name_or_id, broker_id=cluster_id, project_id=project_id)
    print_info_table(res, exclude=["plan"], as_json=as_json)
    return res


@cli.subcommand(
    "actions",
    "logs",
    name_or_id=Arg(help=Messages.name_or_id_optional.format(noun="action")),
    project_id=Arg(help=Messages.project_id.format(noun="action")),
    cluster_id=Arg(help=Messages.cluster_id.format(noun="action")),
    as_json=Arg("--json", help=Messages.as_json),
)
def logs(
    name_or_id: ty.Optional[ty.StrOrUUID] = None,
    project_id: ty.Optional[ty.UUID] = None,
    cluster_id: ty.Optional[ty.UUID] = None,
    as_json: bool = False,
) -> ty.Optional[str]:
    """Get logs for an action on the cluster"""
    job = resolve_action(name_or_id, broker_id=cluster_id, project_id=project_id)
    auth = get_current_auth()
    try:
        res = auth.broker_client.jobs.logs(JobLogRequest(id=job.id))
    except BrokerError as e:
        raise CLIError(Messages.E011.format(noun="action"), e)
    text = res.text if res is not None else None
    print_logs(text, as_json=as_json)
    return text


@cli.subcommand(
    "actions",
    "start",
    name_or_id=Arg(help=Messages.name_or_id_optional.format(noun="action")),
    project_id=Arg(help=Messages.project_id.format(noun="action")),
    cluster_id=Arg(help=Messages.cluster_id.format(noun="action")),
)
def start(
    name_or_id: ty.Optional[ty.StrOrUUID] = None,
    project_id: ty.Optional[ty.UUID] = None,
    cluster_id: ty.Optional[ty.UUID] = None,
) -> ty.UUID:
    """Start an action on the cluster"""
    job = resolve_action(name_or_id, broker_id=cluster_id, project_id=project_id)
    auth = get_current_auth()
    start_job(job, auth)
    return job.id


@cli.subcommand(
    "actions",
    "stop",
    name_or_id=Arg(help=Messages.name_or_id_optional.format(noun="action")),
    project_id=Arg(help=Messages.project_id.format(noun="action")),
    cluster_id=Arg(help=Messages.cluster_id.format(noun="action")),
)
def stop(
    name_or_id: ty.Optional[ty.StrOrUUID] = None,
    project_id: ty.Optional[ty.UUID] = None,
    cluster_id: ty.Optional[ty.UUID] = None,
) -> ty.UUID:
    """Stop an action on the cluster"""
    job = resolve_action(name_or_id, broker_id=cluster_id, project_id=project_id)
    auth = get_current_auth()
    stop_job(job, auth)
    return job.id


@cli.subcommand(
    "actions",
    "delete",
    name_or_id=Arg(help=Messages.name_or_id.format(noun="action")),
    project_id=Arg(help=Messages.project_id.format(noun="action")),
    cluster_id=Arg(help=Messages.cluster_id.format(noun="action")),
)
def delete(
    name_or_id: ty.StrOrUUID,
    project_id: ty.Optional[ty.UUID] = None,
    cluster_id: ty.Optional[ty.UUID] = None,
) -> ty.UUID:
    """Delete an Action by name or ID"""
    job = resolve_action(name_or_id, broker_id=cluster_id, project_id=project_id)
    auth = get_current_auth()
    delete_job(job, auth)
    return job.id



import time
from uuid import UUID

from wasabi import msg

from . import ty
from .auth import CLIAuth, get_current_auth
from .config import SavedSettings
from .errors import CLIError, ProdigyTeamsErrors
from .messages import Messages
from .prodigy_teams_broker_sdk import Client as BrokerClient
from .prodigy_teams_broker_sdk import models as broker_models
from .prodigy_teams_pam_sdk import Client as PamClient
from .prodigy_teams_pam_sdk import models as pam_models
from .prodigy_teams_pam_sdk.client.action import Action as PamActionEndpoint
from .prodigy_teams_pam_sdk.client.task import Task as PamTaskEndpoint
from .prodigy_teams_pam_sdk.errors import PRODIGY_TEAMS_ERRORS, ProdigyTeamsError


def get_not_found_error(model_name: str) -> ty.Type[ProdigyTeamsError]:
    name = model_name.title().replace("_", "")
    return PRODIGY_TEAMS_ERRORS[f"{name}NotFound"]


def resolve_action(
    name_or_id: ty.Optional[ty.Union[str, ty.UUID]],
    *,
    broker_id: ty.Optional[ty.UUID],
    project_id: ty.Optional[ty.UUID],
) -> pam_models.ActionReturning:
    if name_or_id is None:
        settings = SavedSettings.load()
        return settings.action  # type: ignore
    return _resolve_object(
        "action", name_or_id, {"broker_id": broker_id, "project_id": project_id}
    )  # type: ignore


def resolve_action_id(
    name_or_id: ty.Optional[ty.Union[str, ty.UUID]],
    *,
    broker_id: ty.Optional[ty.UUID],
    project_id: ty.Optional[ty.UUID],
) -> ty.UUID:
    if name_or_id is None:
        settings = SavedSettings.load()
        return settings.action  # type: ignore
    return _resolve_id(
        "action", name_or_id, {"broker_id": broker_id, "project_id": project_id}
    )  # type: ignore


def resolve_asset(
    name_or_id: ty.Union[str, ty.UUID],
    *,
    broker_id: ty.Optional[ty.UUID],
    project_id: ty.Optional[ty.UUID],
) -> pam_models.AssetReturning:
    return _resolve_object(
        "asset", name_or_id, {"broker_id": broker_id, "project_id": project_id}
    )  # type: ignore


def resolve_asset_id(
    name_or_id: ty.Union[str, ty.UUID],
    *,
    broker_id: ty.Optional[ty.UUID],
    project_id: ty.Optional[ty.UUID],
) -> ty.UUID:
    return _resolve_id(
        "asset", name_or_id, {"broker_id": broker_id, "project_id": project_id}
    )


def resolve_broker(
    name_or_id: ty.Union[str, ty.UUID],
) -> pam_models.BrokerReturning:
    return _resolve_object("broker", name_or_id, {})  # type: ignore


def resolve_broker_id(
    name_or_id: ty.Union[str, ty.UUID],
) -> ty.UUID:
    return _resolve_id("broker", name_or_id, {})


def resolve_dataset(
    name_or_id: ty.Union[str, ty.UUID],
    *,
    broker_id: ty.Optional[ty.UUID],
) -> pam_models.DatasetReturning:
    return _resolve_object("dataset", name_or_id, {"broker_id": broker_id})  # type: ignore


def resolve_dataset_id(
    name_or_id: ty.Union[str, ty.UUID],
    *,
    broker_id: ty.Optional[ty.UUID],
) -> ty.UUID:
    return _resolve_id("dataset", name_or_id, {"broker_id": broker_id})


def resolve_package(
    name_or_id: ty.Union[str, ty.UUID],
    *,
    broker_id: ty.Optional[ty.UUID],
) -> pam_models.PackageReturning:
    return _resolve_latest_object(
        "package",
        name_or_id,
        {"broker_id": broker_id},
        pam_models.PackageReadingLatest,
    )  # type: ignore


def resolve_package_id(
    name_or_id: ty.Union[str, ty.UUID],
    *,
    broker_id: ty.Optional[ty.UUID],
) -> ty.UUID:
    return _resolve_id("package", name_or_id, {"broker_id": broker_id})


def resolve_path(
    name_or_id: ty.Union[str, ty.UUID],
    *,
    broker_id: ty.Optional[ty.UUID],
) -> pam_models.BrokerPathReturning:
    return _resolve_object("broker_path", name_or_id, {"broker_id": broker_id})  # type: ignore


def resolve_path_id(
    name_or_id: ty.Union[str, ty.UUID],
    *,
    broker_id: ty.Optional[ty.UUID],
) -> ty.UUID:
    return _resolve_id("broker_path", name_or_id, {"broker_id": broker_id})


def resolve_project(
    name_or_id: ty.Union[str, ty.UUID],
) -> pam_models.ProjectReturning:
    return _resolve_object("project", name_or_id, {})  # type: ignore


def resolve_project_id(
    name_or_id: ty.Union[str, ty.UUID],
) -> ty.UUID:
    return _resolve_id("project", name_or_id, {})


def resolve_recipe(
    name_or_id: ty.Union[str, ty.UUID],
    *,
    broker_id: ty.Optional[ty.UUID],
) -> pam_models.RecipeDetail:
    return _resolve_latest_object(
        "recipe",
        name_or_id,
        {"broker_id": broker_id},
        pam_models.RecipeReadingLatest,
    )  # type: ignore


# TODO: this is currently not used
def resolve_recipe_id(
    name_or_id: ty.Union[str, ty.UUID],
    *,
    broker_id: ty.Optional[ty.UUID],
) -> ty.UUID:
    return _resolve_id("recipe", name_or_id, {"broker_id": broker_id})


def resolve_secret(
    name_or_id: ty.Union[str, ty.UUID],
    *,
    broker_id: ty.Optional[ty.UUID],
) -> pam_models.SecretReturning:
    return _resolve_object("secret", name_or_id, {"broker_id": broker_id})  # type: ignore


def resolve_secret_id(
    name_or_id: ty.Union[str, ty.UUID],
    *,
    broker_id: ty.Optional[ty.UUID],
) -> ty.UUID:
    return _resolve_id("secret", name_or_id, {"broker_id": broker_id})


def resolve_task(
    name_or_id: ty.Optional[ty.Union[str, ty.UUID]],
    *,
    broker_id: ty.Optional[ty.UUID],
    project_id: ty.Optional[ty.UUID],
) -> pam_models.TaskReturning:
    return _resolve_object("task", name_or_id, {"broker_id": broker_id, "project_id": project_id})  # type: ignore


def resolve_task_id(
    name_or_id: ty.Optional[ty.Union[str, ty.UUID]],
    *,
    broker_id: ty.Optional[ty.UUID],
    project_id: ty.Optional[ty.UUID],
) -> ty.UUID:
    return _resolve_id("task", name_or_id, {"broker_id": broker_id, "project_id": project_id})  # type: ignore


def _resolve_id(
    model_name: str,
    name_or_id: ty.Union[str, ty.UUID],
    params: ty.Dict[str, ty.Optional[ty.Union[str, ty.UUID]]],
) -> ty.UUID:
    """Resolve reference to a Prodigy Teams Entity by name or ID.
    If an ID is given, return it. If it's a name,
    look up the ID.
    """
    if isinstance(name_or_id, ty.UUID):
        return name_or_id
    try:
        return UUID(name_or_id)
    except ValueError:
        pass
    auth = get_current_auth()
    model_client = getattr(auth.client, model_name)
    NotFound = get_not_found_error(model_name)
    try:
        res = model_client.read(name=name_or_id, **params)
        return res.id
    except NotFound:
        err = Messages.E038.format(noun=model_name, name_or_id=name_or_id)
        raise CLIError(err, params)


def _resolve_object(
    model_name: str,
    name_or_id: ty.Union[str, ty.UUID],
    params: ty.Dict[str, ty.Optional[ty.Union[str, ty.UUID]]],
) -> ty.BaseModel:
    """Resolve reference to a Prodigy Teams Entity by name or ID.
    If an ID is given, use it for an id query, otherwise, query by
    the entity name
    """
    auth = get_current_auth()
    model_client = getattr(auth.client, model_name)
    NotFound = get_not_found_error(model_name)
    record_id = None
    if isinstance(name_or_id, ty.UUID):
        record_id = name_or_id
    else:
        try:
            record_id = UUID(name_or_id)
        except ValueError:
            pass
    if record_id:
        try:
            res = model_client.read(id=record_id, **params)
        except NotFound:
            err = Messages.E038.format(noun=model_name, name_or_id=name_or_id)
            raise CLIError(err, params)
    else:
        try:
            res = model_client.read(name=name_or_id, **params)
        except NotFound:
            err = Messages.E038.format(noun=model_name, name_or_id=name_or_id)
            raise CLIError(err, params)
    return res


def _resolve_latest_object(
    model_name: str,
    name_or_id: ty.Union[str, ty.UUID],
    params: ty.Dict[str, ty.Optional[ty.Union[str, ty.UUID]]],
    latest_request_model: ty.Type[pam_models.BaseModel],
) -> pam_models.BaseModel:
    """Resolve latest reference to a Prodigy Teams Entity by name or ID.
    If an ID is given, use it for an id query, otherwise, query by
    the entity name
    """
    auth = get_current_auth()
    model_client = getattr(auth.client, model_name)
    NotFound = get_not_found_error(model_name)
    record_id = None
    if isinstance(name_or_id, ty.UUID):
        record_id = name_or_id
    else:
        try:
            record_id = UUID(name_or_id)
        except ValueError:
            pass
    if record_id:
        try:
            res = model_client.read(id=record_id, **params)
        except NotFound:
            err = Messages.E038.format(noun=model_name, name_or_id=name_or_id)
            raise CLIError(err, params)
    else:
        body = latest_request_model(name=name_or_id, **params)
        try:
            res = model_client.latest(body=body)
        except NotFound:
            err = Messages.E038.format(
                noun=model_name, name_or_id=getattr(body, "name", None)
            )
            raise CLIError(err, body)
    return res


class JobOperations:
    def __init__(
        self,
        pam_client: PamClient,
        broker_client: BrokerClient,
        job_type: ty.Literal["task", "action"],
    ) -> None:
        self.pam_client = pam_client
        self.broker_client = broker_client
        self.job_type = broker_models.JobType(job_type)

    def start(self, job_id: ty.UUID) -> None:
        change_response = self.broker_client.jobs.start_job(
            broker_models.JobID(id=job_id, job_type=self.job_type)
        )
        if change_response.nomad_change is not None:
            return
        elif change_response.nomad_error:
            err = Messages.E025.format(noun=self.job_type.value, name=job_id)
            raise CLIError(err, change_response.nomad_error)
        elif not change_response.pam_error and not change_response.validation_error:
            msg.info(Messages.T006.format(noun=self.job_type.value, name=job_id))

    def stop(self, job_id: ty.UUID) -> None:
        change_response = self.broker_client.jobs.stop_job(
            broker_models.JobID(id=job_id, job_type=self.job_type)
        )
        if change_response.nomad_change is not None:
            msg.good(Messages.T007.format(noun=self.job_type.value, name=job_id))
        elif change_response.nomad_error:
            err = Messages.E026.format(noun=self.job_type.value, name=job_id)
            raise CLIError(err, change_response.nomad_error)
        elif not change_response.pam_error and not change_response.validation_error:
            msg.info(Messages.T008.format(noun=self.job_type.value, name=job_id))

    def delete(self, job_id: ty.UUID) -> None:
        change_response = self.broker_client.jobs.stop_job(
            broker_models.JobID(id=job_id, job_type=self.job_type)
        )
        if change_response.pam_error is not None:
            raise CLIError(Messages.E027.format(noun=self.job_type.value, name=job_id))
        elif change_response.nomad_error:
            err = Messages.E028.format(noun=self.job_type.value, name=job_id)
            raise CLIError(err, change_response.nomad_error)
        else:
            msg.good(Messages.T009.format(noun=self.job_type.value, name=job_id))
        try:
            self.pam_endpoint().delete(id=job_id)
        except (ProdigyTeamsErrors.TaskNotFound, ProdigyTeamsErrors.ActionNotFound):
            raise CLIError(Messages.E029.format(noun=self.job_type.value, name=job_id))
        except (
            ProdigyTeamsErrors.TaskForbiddenDelete,
            ProdigyTeamsErrors.TaskForbiddenRead,
            ProdigyTeamsErrors.ActionForbiddenDelete,
            ProdigyTeamsErrors.ActionForbiddenRead,
        ):
            raise CLIError(Messages.E030.format(noun=self.job_type.value, name=job_id))
        msg.good(Messages.T003.format(noun=self.job_type.value, name=job_id))

    def pam_endpoint(self) -> ty.Union[PamActionEndpoint, PamTaskEndpoint]:
        return getattr(self.pam_client, self.job_type.value)


_JobT = ty.TypeVar(
    "_JobT", bound=ty.Union[pam_models.TaskReturning, pam_models.ActionReturning]
)


def start_job(job: _JobT, auth: CLIAuth) -> _JobT:
    job_type = "task" if isinstance(job, pam_models.TaskReturning) else "action"
    job_operations = JobOperations(auth.client, auth.broker_client, job_type)
    job_operations.start(job.id)
    SavedSettings.get_and_save(job_type, job.id)
    res = check_job_started(job)
    if res is not None:
        print(res.url)
    return job


def stop_job(job: _JobT, auth: CLIAuth) -> _JobT:
    job_type = "task" if isinstance(job, pam_models.TaskReturning) else "action"
    job_operations = JobOperations(auth.client, auth.broker_client, job_type)
    job_operations.stop(job.id)
    SavedSettings.get_and_save(job_type, job.id)
    return job


def delete_job(job: _JobT, auth: CLIAuth) -> _JobT:
    job_type = "task" if isinstance(job, pam_models.TaskReturning) else "action"
    job_operations = JobOperations(auth.client, auth.broker_client, job_type)
    job_operations.delete(job.id)
    SavedSettings.get_and_save(job_type, None)
    return job


def check_job_started(job: _JobT, *, wait_seconds: int = 60 * 2) -> ty.Optional[_JobT]:
    res = job
    if res.is_running:
        return job
    job_type = "task" if isinstance(job, pam_models.TaskReturning) else "action"
    failed_states = [pam_models.TaskState.FAILED, pam_models.ActionState.FAILED]
    resolver = resolve_task if job_type == "task" else resolve_action
    with msg.loading(Messages.T034.format(noun=job_type)):
        for _ in range(wait_seconds):
            res = resolver(job.id, broker_id=job.broker_id, project_id=job.project_id)
            if res.is_running or res.error is not None or res.state in failed_states:
                break
            time.sleep(1)
    if res.is_running:
        msg.good(Messages.T005.format(noun=job_type, name=job.id))
        return res
    elif res.error is not None or res.state in failed_states:
        raise CLIError(Messages.E025.format(noun=job_type, name=job.id), res.url_logs)
    msg.warn(Messages.E025.format(noun=job_type, name=job.id), res.url_logs)


def collect_from_pages(request: ty.Callable[[int], ty.Page]) -> ty.Iterable[ty.Page]:
    page = 1
    while True:
        page_result = request(page)

        if page_result.items:
            yield page_result
        else:
            break

        page += 1



from ..models import (
    RecipeCreating,
    RecipeDeleting,
    RecipeDetail,
    RecipeListingLatest,
    RecipeReading,
    RecipeReadingLatest,
    RecipeReturning,
    RecipeUpdating,
)
from .base import VersionedModelClient


class Recipe(
    VersionedModelClient[
        RecipeCreating,
        RecipeReading,
        RecipeUpdating,
        RecipeDeleting,
        RecipeReturning,
        RecipeDetail,
        RecipeReadingLatest,
        RecipeListingLatest,
    ]
):
    Creating = RecipeCreating
    Reading = RecipeReading
    Updating = RecipeUpdating
    Deleting = RecipeDeleting
    Returning = RecipeReturning
    Detail = RecipeDetail
    ReadingLatest = RecipeReadingLatest
    ListingLatest = RecipeListingLatest
